# Copyright 2024 CrackNuts. All rights reserved.

import re
import struct
import warnings
from enum import Enum

from cracknuts.cracker import protocol, serial
from cracknuts.cracker.cracker_basic import ConfigBasic, CrackerBasic, connection_status_check


class ConfigS1(ConfigBasic):
    def __init__(self):
        super().__init__()

        self.nut_enable: bool = False
        self.nut_voltage: float = 3.3
        self.nut_clock_enable = False
        self.nut_clock = 8000
        self.nut_timeout: int | None = None
        self.nut_reset_io_enable: bool = False

        self.nut_uart_enable: bool = False
        self.nut_uart_baudrate: serial.Baudrate = serial.Baudrate.BAUDRATE_115200
        self.nut_uart_bytesize: serial.Bytesize = serial.Bytesize.EIGHTBITS
        self.nut_uart_parity: serial.Parity = serial.Parity.PARITY_NONE
        self.nut_uart_stopbits: serial.Stopbits = serial.Stopbits.STOPBITS_ONE

        self.nut_spi_enable: bool | None = False
        self.nut_spi_speed: float = 10_000.0
        self.nut_spi_cpol: serial.SpiCpol = serial.SpiCpol.SPI_CPOL_LOW
        self.nut_spi_cpha: serial.SpiCpha = serial.SpiCpha.SPI_CPHA_LOW
        self.nut_spi_csn_auto: bool = True
        self.nut_spi_csn_delay: bool = True

        self.nut_i2c_enable: bool = False
        self.nut_i2c_dev_addr: int = 0x00
        self.nut_i2c_speed: serial.I2cSpeed = serial.I2cSpeed.STANDARD_100K
        self.nut_i2c_stretch_enable = False

    def __str__(self):
        return super().__str__()

    def __repr__(self):
        return super().__repr__()


class CrackerS1(CrackerBasic[ConfigS1]):
    def get_default_config(self) -> ConfigS1:
        return ConfigS1()

    @connection_status_check
    def get_current_config(self) -> ConfigS1 | None:
        status, res = self.send_with_command(protocol.Command.GET_CONFIG)
        if status == protocol.STATUS_OK:
            # === Since the device does not support the channel enable function,
            # the information is temporarily saved to the host software. ===
            # return self._parse_config_bytes(res)
            config = self._parse_config_bytes(res)
            config.osc_channel_0_enable = self._channel_enable.osc_channel_0_enable
            config.osc_channel_1_enable = self._channel_enable.osc_channel_1_enable
            # Cracker only supports sampling length in multiples of 1024,
            # record actual length to truncate waveform data later.
            config.osc_sample_length = self._osc_sample_length
            return config
            # === end ===
        else:
            self._logger.error(f"Get config from cracker error, return code 0x{status:02x}.")
            return None

    def _update_config_from_cracker(self):
        for k, v in self.get_current_config().__dict__.items():
            setattr(self._config, k, v)

    def _parse_config_bytes(self, config_bytes: bytes) -> ConfigS1 | None:
        if config_bytes is None:
            return None
        bytes_format, config = self._get_config_bytes_format()
        res_bytes_length = len(config_bytes)
        struct_fmt = f">{"".join(bytes_format.values())}"
        bytes_length = struct.calcsize(struct_fmt)
        if res_bytes_length != bytes_length:
            if res_bytes_length < bytes_length:
                self._logger.warning(
                    f"The length of the config info from Cracker is incorrect: "
                    f"expected {bytes_length}, but got {res_bytes_length}."
                )
                return None
            config_bytes = config_bytes[:bytes_length]

        try:
            config_tuple = struct.unpack(f">{"".join(bytes_format.values())}", config_bytes)
        except Exception as e:
            self._logger.error(
                f"Parse config bytes error: {e.args}, origin bytes is: [{config_bytes.hex(' ')}], "
                f"Default configuration will be used."
            )
            return config

        for i, k in enumerate(bytes_format.keys()):
            v = config_tuple[i]
            if not hasattr(config, k):
                self._logger.warning(f"Parse config bytes error: {k} is not a valid config key.")
            else:
                default_value = getattr(config, k)
                if k == "nut_voltage":
                    v = v / 1000
                elif k == "nut_spi_speed":
                    v = round(100e6 / 2 / v, 2)
                elif default_value is not None and isinstance(default_value, Enum):
                    v = default_value.__class__(v)

                setattr(config, k, v)

        return config

    def _get_config_bytes_format(self) -> tuple[dict[str, str], ConfigS1]:
        return (
            {
                "nut_enable": "?",
                "nut_voltage": "I",
                "nut_clock_enable": "?",
                "nut_clock": "I",
                "nut_reset_io_enable": "?",
                "osc_sample_clock": "I",
                "osc_sample_phase": "I",
                "osc_sample_length": "I",
                "osc_sample_delay": "I",
                "osc_channel_0_enable": "?",
                "osc_channel_0_gain": "B",
                "osc_channel_1_enable": "?",
                "osc_channel_1_gain": "B",
                "osc_trigger_mode": "B",
                "osc_trigger_source": "B",
                "osc_trigger_edge": "B",
                "osc_trigger_edge_level": "H",
                "nut_i2c_enable": "?",
                "nut_i2c_dev_addr": "B",
                "nut_i2c_speed": "B",
                "nut_i2c_stretch_enable": "?",
                "nut_uart_enable": "?",
                "nut_uart_stopbits": "B",
                "nut_uart_parity": "B",
                "nut_uart_bytesize": "B",
                "nut_uart_baudrate": "B",
                "nut_spi_enable": "?",
                "nut_spi_speed": "H",
                "nut_spi_cpol": "B",
                "nut_spi_cpha": "B",
                "nut_spi_csn_auto": "?",
                "nut_spi_csn_delay": "?",
            },
            ConfigS1(),
        )

    @connection_status_check
    def write_config_to_cracker(self, config: ConfigS1):
        self._osc_set_analog_all_channel_enable({0: config.osc_channel_0_enable, 1: config.osc_channel_1_enable})
        self.osc_analog_gain(0, config.osc_channel_0_gain)
        self.osc_analog_gain(1, config.osc_channel_1_gain)
        self.osc_sample_length(config.osc_sample_length)
        self.osc_sample_delay(config.osc_sample_delay)
        self.osc_sample_clock(config.osc_sample_clock)
        self.osc_sample_clock_phase(config.osc_sample_phase)
        self.osc_trigger_source(config.osc_trigger_source)
        self.osc_trigger_mode(config.osc_trigger_mode)
        self.osc_trigger_edge(config.osc_trigger_edge)
        self.osc_trigger_level(config.osc_trigger_edge_level)

        self.spi_config(
            config.nut_spi_speed,
            config.nut_spi_cpol,
            config.nut_spi_cpha,
            config.nut_spi_csn_auto,
            config.nut_spi_csn_delay,
        )
        self._spi_enable(config.nut_spi_enable)

        self.uart_config(
            config.nut_uart_baudrate,
            config.nut_uart_bytesize,
            config.nut_uart_parity,
            config.nut_uart_stopbits,
        )
        self._uart_io_enable(config.nut_uart_enable)

        self.i2c_config(config.nut_i2c_dev_addr, config.nut_i2c_speed)
        self._i2c_enable(config.nut_i2c_enable)

        self.nut_voltage(config.nut_voltage)
        self._nut_set_enable(config.nut_enable)

        self.nut_clock_freq(config.nut_clock)
        self._nut_set_clock_enable(config.nut_clock_enable)

    @connection_status_check
    def register_read(self, base_address: int, offset: int) -> tuple[int, bytes | None]:
        """
        Read register.

        :param base_address: Base address of the register.
        :type base_address: int
        :param offset: Offset of the register.
        :type offset: int
        :return: The device response status and the value read from the register or None if an exception is raised.
        :rtype: tuple[int, bytes | None]
        """
        payload = struct.pack(">II", base_address, offset)
        self._logger.debug(f"register_read payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.CRACKER_READ_REGISTER, payload=payload)
        if status != protocol.STATUS_OK:
            return status, None
        else:
            return status, res

    @connection_status_check
    def register_write(self, base_address: int, offset: int, data: bytes | int | str) -> tuple[int, None]:
        """
        Write register.

        :param base_address: Base address of the register.
        :type base_address: int
        :param offset: Offset of the register.
        :type offset: int
        :param data: Data to write.
        :return: The device response status
        :rtype: tuple[int, None]
        """
        if isinstance(data, str):
            if data.startswith("0x") or data.startswith("0X"):
                data = data[2:]
            data = bytes.fromhex(data)
        if isinstance(data, int):
            data = struct.pack(">I", data)
        payload = struct.pack(">II", base_address, offset) + data
        self._logger.debug(f"register_write payload: {payload.hex()}")
        status, _ = self.send_with_command(protocol.Command.CRACKER_WRITE_REGISTER, payload=payload)
        return status, None

    def osc_analog_enable(self, channel: int | str) -> tuple[int, None]:
        """
        Enable osc analog.

        :param channel: Channel to enable, it can be 'A' or 'B', or 0, 1
        :type channel: int | str
        :return: The device response status
        :rtype: tuple[int, None]
        """
        return self._osc_set_analog_channel_enable(channel, True)

    def osc_analog_disable(self, channel: int | str) -> tuple[int, None]:
        """
        Disable osc analog.

        :param channel: Channel to enable, it can be 'A' or 'B', or 0, 1
        :type channel: int | str
        :return: The device response status
        :rtype: tuple[int, None]
        """
        return self._osc_set_analog_channel_enable(channel, False)

    @connection_status_check
    def _osc_set_analog_channel_enable(self, channel: int | str, enable: bool) -> tuple[int, None]:
        """
        Set analog channel enable.

        :param channel: Channel to enable, it can be 'A' or 'B', or 0, 1
        :type channel: int | str
        :param enable: Enable or disable.
        :type enable: bool
        :return: The device response status
        :rtype: tuple[int, None]
        """

        # === Since the device does not support the channel enable function,
        # the information is temporarily saved to the host software. ===
        # config = self.get_current_config()
        # === end ===
        config = self._channel_enable

        channels = ("A", "B")
        if isinstance(channel, str):
            channel = channel.upper()
            if channel in channels:
                channel = channels.index(channel)
            else:
                self._logger.error("Channel error, it should in 'A' or 'B'")
                return self.NON_PROTOCOL_ERROR, None
        else:
            if channel > 1:
                self._logger.error("Channel error, it should in 0 or 1")
                return self.NON_PROTOCOL_ERROR, None
        if channel == 0:
            final_enable = {0: enable, 1: config.osc_channel_1_enable}
        elif channel == 1:
            final_enable = {0: config.osc_channel_0_enable, 1: enable}
        else:
            final_enable = {0: config.osc_channel_0_enable, 1: config.osc_channel_1_enable}
        # === Since the device does not support the channel enable function,
        # the information is temporarily saved to the host software. ===
        # status, res = self._osc_set_analog_all_channel_enable(final_enable)
        self._channel_enable.osc_channel_0_enable = final_enable[0]
        self._channel_enable.osc_channel_1_enable = final_enable[1]
        self._config.osc_channel_0_enable = final_enable[0]
        self._config.osc_channel_1_enable = final_enable[1]
        # === end ===

    @connection_status_check
    def _osc_set_analog_all_channel_enable(self, final_enable: dict[int, bool]):
        mask = 0
        if final_enable.get(0):
            mask |= 1
        if final_enable.get(1):
            mask |= 1 << 1
        if final_enable.get(2):
            mask |= 1 << 2
        if final_enable.get(3):
            mask |= 1 << 3
        if final_enable.get(4):
            mask |= 1 << 4
        if final_enable.get(5):
            mask |= 1 << 5
        if final_enable.get(6):
            mask |= 1 << 6
        if final_enable.get(7):
            mask |= 1 << 7
        if final_enable.get(8):
            mask |= 1 << 8
        payload = struct.pack(">B", mask)
        return self.send_with_command(protocol.Command.OSC_ANALOG_CHANNEL_ENABLE, payload=payload)

    @connection_status_check
    def osc_trigger_mode(self, mode: int | str) -> tuple[int, None]:
        """
        Set trigger mode.

        :param mode: Trigger mode. Trigger mode can be one of ("EDGE", "PATTERN") or ("E", "P"),
                     or the index of the string value.
        :type mode: int | str
        :return: The device response status
        :rtype: tuple[int, None]
        """
        modes1 = ("EDGE", "PATTERN")
        modes2 = ("E", "P")
        if isinstance(mode, str):
            mode = mode.upper()
            if mode in modes1:
                mode = modes1.index(mode)
            elif mode in modes2:
                mode = modes2.index(mode)
            else:
                self._logger.error(f"Mode must be one of {modes1} or {modes2} if specified as a string.")
                return self.NON_PROTOCOL_ERROR, None
        else:
            if mode > 1:
                self._logger.error(f"Unknown trigger mode: {mode}, it must be one of (0, 1)")
                return self.NON_PROTOCOL_ERROR, None
        payload = struct.pack(">B", mode)
        self._logger.debug(f"osc_trigger_mode payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.OSC_TRIGGER_MODE, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.osc_trigger_mode = mode

        return status, None

    @connection_status_check
    def osc_trigger_source(self, source: int | str) -> tuple[int, None]:
        """
        Set trigger source.

        :param source: Trigger source: It can be one of ('N', 'A', 'B', 'P') or ('Nut'、'ChA'、'ChB'、'Protocol')
                       or a number in 0, 1, 2, 3, represent Nut, Channel A, Channel B, and Protocol, respectively.
        :type source: int | str
        :return: The device response status
        :rtype: tuple[int, None]
        """
        sources1 = ("N", "A", "B", "P")
        sources2 = ("NUT", "CHA", "CHB", "PROTOCOL")
        if isinstance(source, str):
            source = source.upper()
            if source in sources1:
                source = sources1.index(source)
            elif source in sources2:
                source = sources2.index(source)
            else:
                self._logger.error(
                    f"Invalid trigger source: {source}. it must be one of {sources1} or {sources2} "
                    f"if specified as a string."
                )
                return self.NON_PROTOCOL_ERROR, None
        else:
            if source > 3:
                self._logger.error("Invalid trigger source, it must be one of (0, 1, 2, 3)")
                return self.NON_PROTOCOL_ERROR, None
        payload = struct.pack(">B", source)
        self._logger.debug(f"osc_trigger_source payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.OSC_ANALOG_TRIGGER_SOURCE, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.osc_trigger_source = source

        return status, None

    @connection_status_check
    def osc_trigger_edge(self, edge: int | str) -> tuple[int, None]:
        """
        Set trigger edge.

        :param edge: Trigger edge. ('up', 'down', 'either') or ('u', 'd', 'e') or 0, 1, 2
                     represent up, down, either, respectively.
        :type edge: int | str
        :return: The device response status
        :rtype: tuple[int, None]
        """
        edges1 = ("UP", "DOWN", "EITHER")
        edges2 = ("U", "D", "E")
        if isinstance(edge, str):
            edge = edge.upper()
            if edge in edges1:
                edge = edges1.index(edge)
            elif edge in edges2:
                edge = edges2.index(edge)
            else:
                self._logger.error(
                    f"Invalid trigger edge: {edge}. it must be one of {edges1} or {edges2} "
                    f"if specified as a string."
                )
                return self.NON_PROTOCOL_ERROR, None
        if edge not in (0, 1, 2):
            self._logger.error("Invalid trigger edge, it must be one of (0, 1, 2)")
            return self.NON_PROTOCOL_ERROR, None

        payload = struct.pack(">B", edge)
        self._logger.debug(f"osc_trigger_edge payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.OSC_TRIGGER_EDGE, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.osc_trigger_edge = edge

        return status, None

    def osc_trigger_level(self, edge_level: int) -> tuple[int, None]:
        """
        Set trigger edge level.

        :param edge_level: Edge level.
        :type edge_level: int
        :return: The device response status
        :rtype: tuple[int, None]
        """
        payload = struct.pack(">h", edge_level)
        self._logger.debug(f"osc_trigger_level payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.OSC_TRIGGER_EDGE_LEVEL, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.osc_trigger_edge_level = edge_level

        return status, None

    @connection_status_check
    def osc_sample_delay(self, delay: int | str) -> tuple[int, None]:
        """
        Set sample delay.

        :param delay: Sample delay. It can be a number or a string, if the input is a string,
                       you can specify the unit, ending with the value,
                       using either 'k' (representing 1024) or 'm' (representing 1024^2).
                       For example, "1k" means 1024 and "2m" means 2097152.
        :type delay: int
        :return: The device response status
        :rtype: tuple[int, None]
        """
        if isinstance(delay, str):
            m = re.match(r"^(\d+)([kKmM])?$", delay)

            if m:
                delay = int(m.group(1))
            else:
                self._logger.error(
                    "Input format error, it should be a number or a string "
                    "ending with the value, using either k or m."
                )
                return self.NON_PROTOCOL_ERROR, None

            if m.group(2):
                unit = m.group(2).lower()
                if unit == "k":
                    delay *= 1024
                elif unit == "m":
                    delay *= 1024 * 1024
                else:
                    self._logger.error(f"Unknown unit: {unit}")
                    return self.NON_PROTOCOL_ERROR, None
        payload = struct.pack(">i", delay)
        self._logger.debug(f"osc_sample_delay payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.OSC_SAMPLE_DELAY, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.osc_sample_delay = delay

        return status, None

    @connection_status_check
    def osc_sample_length(self, length: int | str) -> tuple[int, None]:
        """
        Set sample length.

        :param length: Sample length. It can be a number or a string, if the input is a string,
                       you can specify the unit, ending with the value,
                       using either 'k' (representing 1024) or 'm' (representing 1024^2).
                       For example, "1k" means 1024 and "2m" means 2097152.
        :type length: int | str
        :return: The device response status
        :rtype: tuple[int, None]
        """
        if isinstance(length, str):
            m = re.match(r"^(\d+)([kKmM])?$", length)

            if m:
                length = int(m.group(1))
            else:
                self._logger.error(
                    "Input format error, it should be a number or a string "
                    "ending with the value, using either k or m."
                )
                return self.NON_PROTOCOL_ERROR, None

            if m.group(2):
                unit = m.group(2).lower()
                if unit == "k":
                    length *= 1024
                elif unit == "m":
                    length *= 1024 * 1024
                else:
                    self._logger.error(f"Unknown unit: {unit}")
                    return self.NON_PROTOCOL_ERROR, None
        # Cracker only supports sampling length in multiples of 1024,
        # record actual length to truncate waveform data later.
        self._osc_sample_length = length
        payload = struct.pack(">I", length)
        self._logger.debug(f"osc_sample_length payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.OSC_SAMPLE_LENGTH, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.osc_sample_length = length

        return status, None

    @connection_status_check
    def osc_sample_clock(self, clock: int | str) -> tuple[int, None]:
        """
        Set osc sample rate

        :param clock: The sample rate in kHz can be one of (65000, 48000, 24000, 12000, 8000)
                      or a string in (65M, 48M, 24M, 12M, 8M).
        :type clock: int | str
        :return: The device response status
        :rtype: tuple[int, None]
        """
        if isinstance(clock, str):
            clock = clock.upper()
            if clock == "65M":
                clock = 65000
            elif clock == "48M":
                clock = 48000
            elif clock == "24M":
                clock = 24000
            elif clock == "12M":
                clock = 12000
            elif clock == "8M":
                clock = 8000
            else:
                if re.match(r"^\d+$", clock):
                    clock = int(clock)
                else:
                    self._logger.error("UnSupport osc sample rate, it should in 65M or 48M or 24M or 12M or 8M")
                    return self.NON_PROTOCOL_ERROR, None
        if clock not in (65000, 48000, 24000, 12000, 8000, 4000):
            self._logger.error("UnSupport osc sample clock, it should in (65000, 48000, 24000, 12000, 8000)")
            return self.NON_PROTOCOL_ERROR, None
        payload = struct.pack(">I", clock)
        self._logger.debug(f"osc_sample_clock_rate payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.OSC_SAMPLE_RATE, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.osc_sample_clock = clock

        return status, None

    @connection_status_check
    def osc_analog_gain(self, channel: int | str, gain: int) -> tuple[int, None]:
        """
        Set analog gain.

        :param channel: Analog channel, it can be 'A' or 'B', or 0, 1
        :type channel: int | str
        :param gain: Analog gain.
        :type gain: int
        :return: The device response status
        :rtype: tuple[int, None]
        """
        channels = ("A", "B")
        if isinstance(channel, str):
            channel = channel.upper()
            if channel in channels:
                channel = channels.index(channel)
            else:
                self._logger.error("Channel error, it should in 'A' or 'B'")
                return self.NON_PROTOCOL_ERROR, None
        else:
            if channel > 1:
                self._logger.error("Channel error, it should in 0 or 1")
                return self.NON_PROTOCOL_ERROR, None

        if not 1 <= gain <= 50:
            self._logger.error("Gain error, it should in 1 to 50")
            return self.NON_PROTOCOL_ERROR, None
        payload = struct.pack(">BB", channel, gain)
        status, res = self.send_with_command(protocol.Command.OSC_ANALOG_GAIN, payload=payload)
        if status == protocol.STATUS_OK:
            if channel == 0:
                self._config.osc_channel_0_gain = gain
            elif channel == 1:
                self._config.osc_channel_1_gain = gain

    @connection_status_check
    def osc_sample_clock_phase(self, phase: int) -> tuple[int, None]:
        """
        Set sample phase.

        :param phase: Sample phase.
        :type phase: int
        :return: The device response status
        :rtype: tuple[int, None]
        """
        payload = struct.pack(">I", phase)
        self._logger.debug(f"osc_sample_clock_phase payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.OSC_CLOCK_SAMPLE_PHASE, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.osc_sample_phase = phase

        return status, None

    def nut_voltage_disable(self) -> tuple[int, None]:
        """
        Disable the nut voltage.

        :return: The device response status
        :rtype: tuple[int, None]
        """
        return self._nut_set_enable(False)

    def nut_voltage_enable(self) -> tuple[int, None]:
        """
        Enable the nut voltage.

        :return: The device response status
        :rtype: tuple[int, None]
        """
        return self._nut_set_enable(True)

    @connection_status_check
    def _nut_set_enable(self, enable: int | bool) -> tuple[int, None]:
        """
        Set nut enable.

        :param enable: Enable or disable. 0 for disable, 1 for enable if specified as a integer.
        :type enable: int | bool
        :return: The device response status
        :rtype: tuple[int, None]
        """
        if isinstance(enable, int):
            enable = True if enable == 1 else False
        payload = struct.pack(">?", enable)
        self._logger.debug(f"nut_enable payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.NUT_ENABLE, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.nut_enable = enable

        return status, None

    @connection_status_check
    def nut_voltage(self, voltage: float | str | int) -> tuple[int, None]:
        """
        Set nut voltage.

        :param voltage: Nut voltage can be a number or a string. If the input is a float,
                        the voltage will be treated as a V value. If the input is a string,
                        you can specify the unit, ending with the value, using either mV or V.
        :type voltage: float | str | int
        :return: The device response status
        :rtype: tuple[int, None]
        """
        if isinstance(voltage, str):
            m = re.match(r"^(\d+(?:\.\d+)?)([mM]?[vV])?$", voltage)
            if not m:
                self._logger.error(
                    "Input format error: " "it should be a number or a string and end with a unit in V or mV."
                )
                return self.NON_PROTOCOL_ERROR, None
            voltage = float(m.group(1))
            if m.group(2):
                unit = m.group(2)
            else:
                unit = "v"
            if unit.lower() == "v":
                voltage = voltage * 1000
        else:
            voltage = voltage * 1000
        voltage = int(voltage)
        if voltage > 4000 or voltage < 2000:
            self._logger.error("Nut voltage error, it should in 2.0V to 4.0V")
            return self.NON_PROTOCOL_ERROR, None
        payload = struct.pack(">I", voltage)
        self._logger.debug(f"nut_voltage payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.NUT_VOLTAGE, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.nut_voltage = voltage / 1000

        return status, None

    @connection_status_check
    def nut_clock_disable(self) -> tuple[int, None]:
        """
        Disable the nut clock.

        :return: The device response status
        :rtype: tuple[int, None]
        """
        return self._nut_set_clock_enable(False)

    @connection_status_check
    def nut_clock_enable(self) -> tuple[int, None]:
        """
        Enable the nut clock.

        :return: The device response status
        :rtype: tuple[int, None]
        """
        return self._nut_set_clock_enable(True)

    @connection_status_check
    def _nut_set_clock_enable(self, enable: bool) -> tuple[int, None]:
        payload = struct.pack(">?", enable)
        self._logger.debug(f"nut_set_clock_enable payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.NUT_CLOCK_ENABLE, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.nut_clock_enable = enable

        return status, None

    @connection_status_check
    def nut_clock_freq(self, clock: int | str) -> tuple[int, None]:
        """
        Set nut clock.

        :param clock: The clock of the nut in kHz
        :type clock: int | str
        :return: The device response status
        :rtype: tuple[int, None]
        """
        if isinstance(clock, str):
            clock = clock.upper()
            if clock == "24M":
                clock = 24000
            elif clock == "12M":
                clock = 12000
            elif clock == "8M":
                clock = 8000
            elif clock == "4M":
                clock = 4000
            else:
                self._logger.error(f"Unknown clock type: {clock}, 24M or 12M or 8M or 4M")
                return protocol.STATUS_ERROR, None

        payload = struct.pack(">I", clock)
        self._logger.debug(f"nut_set_clock payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.NUT_CLOCK, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.nut_clock = clock

        return status, None

    @connection_status_check
    def nut_timeout_ms(self, timeout: int) -> tuple[int, None]:
        """
        Set nut timeout.

        :param timeout: Nut timeout.
        :type timeout: int
        :return: The device response status
        :rtype: tuple[int, None]
        """
        payload = struct.pack(">I", timeout)
        self._logger.debug(f"nut_timeout_ms payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.NUT_TIMEOUT, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.nut_timeout = timeout

        return status, None

    def spi_enable(self) -> tuple[int, None]:
        """
        启用SPI

        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, None]
        """
        return self._spi_enable(True)

    def spi_disable(self) -> tuple[int, None]:
        """
        禁用 SPI

        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, None]
        """
        return self._spi_enable(False)

    @connection_status_check
    def _spi_enable(self, enable: bool):
        """
        配置SPI接口是否使能

        :param enable: True 启用, False 禁用.
        :type enable: bool
        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, None]
        """
        payload = struct.pack(">?", enable)
        self._logger.debug(f"cracker_spi_enable payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.CRACKER_SPI_ENABLE, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.nut_spi_enable = enable
        return status, res

    @connection_status_check
    def spi_reset(self) -> tuple[int, None]:
        """
        复位SPI硬件。

        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, None]
        """
        payload = None
        self._logger.debug(f"cracker_spi_reset payload: {payload}")
        return self.send_with_command(protocol.Command.CRACKER_SPI_RESET)

    @connection_status_check
    def spi_config(
        self,
        speed: float | None = None,
        cpol: serial.SpiCpol | None = None,
        cpha: serial.SpiCpha | None = None,
        csn_auto: bool | None = None,
        csn_delay: bool | None = None,
    ) -> tuple[int, None | str]:
        """
        配置SPI接口参数。CPOL（时钟极性）和CPHA（时钟相位）

        :param speed: 通信速率，默认10kHz，同CFG_PROTOCOL寄存器中PSC关系为：speed=(100×10^6)/(2*PSC)
        :type speed: int
        :param cpol: 时钟极性.
        :type cpol: serial.SpiCpol
        :param cpha: 时钟相位.
        :type cpha: serial.SpiCpha
        :param csn_auto: True，片选（Chip Select）信号只在数据通信时拉低。False，片选信号一直拉低。
        :type csn_auto: bool,
        :param csn_delay: True，片选（Chip Select）信号在Delay过程中保持低电平。False，片选信号在Delay期间为高电平。
        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, None]
        """

        if speed is None or cpol is None or csn_auto is None or csn_delay is None:
            config = self.get_current_config()
            if config is None:
                self._logger.error("Get config from cracker error.")
                return self.NON_PROTOCOL_ERROR, None
            if speed is None:
                speed = config.nut_spi_speed
            if cpol is None:
                cpol = config.nut_spi_cpol
            if cpha is None:
                cpha = config.nut_spi_cpha
            if csn_auto is None:
                csn_auto = config.nut_spi_csn_auto
            if csn_delay is None:
                csn_delay = config.nut_spi_csn_delay

        # System clock is 100e6
        # Clock divider is 2
        # psc max is 65535
        psc = 100e6 / 2 / speed
        if psc > 65535 or psc < 2:
            self._logger.error("Not support speed.")
            return self.NON_PROTOCOL_ERROR, f"Not support speed {speed}."

        if not psc.is_integer():
            psc = round(psc)
            if psc > 65535 or psc < 2:
                self._logger.error(f"Not support  speed {speed}.")
                return self.NON_PROTOCOL_ERROR, f"Not support speed {speed}."
            _speed = speed
            speed = round(100e6 / 2 / psc, 2)
            self._logger.warning(
                f"The speed: [{_speed}] cannot calculate an integer Prescaler, "
                f"so the integer value is set to {speed} and psc is {psc}."
            )

        speed = round(100e6 / 2 / psc, 2)

        payload = struct.pack(">HBB??", int(psc), cpol.value, cpha.value, csn_auto, csn_delay)
        self._logger.debug(f"cracker_spi_config payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.CRACKER_SPI_CONFIG, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.nut_spi_speed = speed
            self._config.nut_spi_cpol = cpol
            self._config.nut_spi_cpha = cpha
            self._config.nut_spi_csn_auto = csn_auto
            self._config.nut_spi_csn_delay = csn_delay
        return status, res

    @staticmethod
    def _find_divisors(clock):
        if clock <= 0:
            raise ValueError("Input must be a positive integer.")

        n = clock / 2

        divisors = []

        for i in range(1, int(n**0.5) + 1):
            q, b = divmod(n, i)
            if b == 0 and 2 < q < 65535:
                divisors.append(int(i))
                if i != n // i:
                    divisors.append(int(n // i))

        return sorted(divisors)

    @connection_status_check
    def spi_transceive_delay_transceive(
        self, tx_data1: bytes | str | None, tx_data2: bytes | str | None, is_delay: bool, delay: int, is_trigger: bool
    ) -> tuple[int, tuple[bytes | None, bytes | None] | None]:
        """
        通过SPI发送tx_data1，等待delay后（单位10ns），读取再发送tx_data2数据。

        is_trigger=True 时，tx_data传输完毕后，Trigger 信号拉高

        ::

            TRIG: ───┐            ┌────────────┐            ┌─── HIGH
                     |            |            |            |
                     └────────────┘            └────────────┘    LOW
                     ┌────────────┬────────────┬────────────┐
                     │  tx_data1  │    delay   │  tx_data2  │
                     └────────────┴────────────┴────────────┘

        is_trigger=False 时，Trigger 信号不变

        ::

            TRIG: ────────────────────────────────────────────── HIGH

                                                                 LOW
                     ┌────────────┬────────────┬────────────┐
                     │  tx_data1  │   timeout  │  rx_data2  │
                     └────────────┴────────────┴────────────┘

        :param tx_data1: 第一阶段待发送数据。
        :type tx_data1: str | bytes
        :param tx_data2: 第二阶段待发送数据。
        :type tx_data2: str | bytes
        :param is_delay: 是否开启delay
        :type is_delay: bool
        :param delay: 发送和接收之间的延时，单位10纳秒。
        :type delay: int
        :param is_trigger: 在数据发送时是否产生触发信息（即：发送开始时拉低，结束时拉高）
        :type is_trigger: bool
        :return: Cracker设备响应状态和接收到的数据：(status, (response1, response2))。
        :rtype: tuple[int, bytes | None]
        """
        if isinstance(tx_data1, str):
            tx_data1 = bytes.fromhex(tx_data1)
        if isinstance(tx_data2, str):
            tx_data2 = bytes.fromhex(tx_data2)
        tx_data1_len = 0 if tx_data1 is None else len(tx_data1)
        tx_data2_len = 0 if tx_data2 is None else len(tx_data2)
        payload = struct.pack(">?IH?", is_delay, delay, tx_data2_len, is_trigger)
        if tx_data1 is not None:
            payload += tx_data1
        if tx_data2 is not None:
            payload += tx_data2
        self._logger.debug(f"spi_transmit_delay_receive payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.CRACKER_SPI_TRANSCEIVE, payload=payload)
        if status != protocol.STATUS_OK:
            self._logger.error(f"Receive status code error [{status}]")
            return status, (None, None)
        else:
            return status, (res[:tx_data1_len], res[-tx_data2_len:])

    def spi_transmit(self, tx_data: bytes | str, is_trigger: bool = False) -> tuple[int, None]:
        """
        通过SPI接口发送数据，并根据is_trigger决定在数据发送后是否产生触发信号。

        is_trigger=True 时，tx_data传输完毕后，Trigger 信号拉高

        ::

            TRIG: ───┐            ┌─── HIGH
                     |            |
                     └────────────┘    LOW
                     ┌────────────┐
                     │   tx_data  │
                     └────────────┘

        is_trigger=False 时，Trigger 信号不变

        ::

            TRIG: ──────────────────── HIGH

                                       LOW
                     ┌────────────┐
                     │   tx_data  │
                     └────────────┘

        :param tx_data: 待发送的数据
        :type tx_data: str | bytes
        :param is_trigger: 在数据发送时是否产生触发信息（即：发送开始时拉低，结束时拉高）
        :type is_trigger: bool
        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, None]
        """
        status, _ = self.spi_transceive_delay_transceive(
            tx_data1=tx_data, tx_data2=None, is_delay=False, delay=1_000_000_000, is_trigger=is_trigger
        )
        return status, None

    def spi_receive(
        self, rx_count: int, dummy: bytes | str = b"\x00", is_trigger: bool = False
    ) -> tuple[int, bytes | None]:
        """
        通过SPI接口读取rx_count个bytes型数据，根据 is_trigger 决定在数据接收后是否产生触发信号。

        is_trigger=True 时，tx_data传输完毕后，Trigger 信号拉高

        ::

            TRIG: ───┐            ┌─── HIGH
                     |            |
                     └────────────┘    LOW
                     ┌────────────┐
                     │   rx_data  │
                     └────────────┘

        is_trigger=False 时，Trigger 信号不变

        ::

            TRIG: ──────────────────── HIGH

                                       LOW
                     ┌────────────┐
                     │   rx_data  │
                     └────────────┘

        :param rx_count: 要读取数据字节长度。
        :type rx_count: int
        :param dummy: 需要在spi读取阶段发送的填充数据
        :type dummy: bytes|str
        :param is_trigger: 在数据发送时是否产生触发信息（即：发送开始时拉低，结束时拉高）
        :type is_trigger: bool
        :return: Cracker设备响应状态和接收到的数据：(status, response)。
                 Return None if an exception is caught.
        :rtype: tuple[int, bytes | None]
        """
        if isinstance(dummy, str):
            dummy = bytes.fromhex(dummy)
        dummy = dummy[:1]

        status, (_, res) = self.spi_transceive_delay_transceive(
            tx_data1=dummy * rx_count, tx_data2=None, is_delay=False, delay=1_000_000_000, is_trigger=is_trigger
        )

        return status, res

    def spi_transmit_delay_receive(
        self, tx_data: bytes | str, delay: int, rx_count: int, dummy: bytes | str = b"\x00", is_trigger: bool = False
    ) -> tuple[int, bytes | None]:
        """
        通过SPI发送bytes型数据tx_data，等待delay后（单位10ns），读取rx_count长度数据。其实现方式是在发送完数据后，
        等待 delay 时间，master 再发送等于 rx_count 长度的dummy数据，master 解析此时 slave 发送来的数据作为 rx_data。


        is_trigger=True 时，tx_data传输开始时 Trigger 信号拉低，完毕后 Trigger 信号拉高

        ::

            TRIG: ───┐            ┌────────────┐            ┌─── HIGH
                     |            |            |            |
                     └────────────┘            └────────────┘    LOW
                     ┌────────────┬────────────┬────────────┐
            MASTER   │  tx_data   │    delay   │    dummy   │
                     └────────────┴────────────┴────────────┘
                     ┌────────────┬────────────┬────────────┐
            SLAVE    │   dummy    │    delay   │   rx_data  │
                     └────────────┴────────────┴────────────┘

        is_trigger=False 时，Trigger 信号不变

        ::

            TRIG: ────────────────────────────────────────────── HIGH

                                                                 LOW
                     ┌────────────┬────────────┬────────────┐
            MASTER   │  tx_data   │    delay   │    dummy   │
                     └────────────┴────────────┴────────────┘
                     ┌────────────┬────────────┬────────────┐
            SLAVE    │   dummy    │    delay   │   rx_data  │
                     └────────────┴────────────┴────────────┘

        :param tx_data: 待发送的数据。
        :type tx_data: str | bytes
        :param delay: 发送和接收之间的延时，单位10纳秒。
        :type delay: int
        :param rx_count: 要读取数据字节长度。
        :type rx_count: int
        :param dummy: 需要在spi读取阶段发送的填充数据
        :type dummy: bytes|str
        :param is_trigger: 在数据发送时是否产生触发信息（即：发送开始时拉低，结束时拉高）
        :type is_trigger: bool
        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, bytes | None]
        """
        status, (_, res) = self.spi_transceive_delay_transceive(
            tx_data1=tx_data, tx_data2=dummy * rx_count, is_delay=True, delay=delay, is_trigger=is_trigger
        )

        return status, res

    def spi_transceive(
        self, tx_data: bytes | str, rx_count: int = None, dummy: bytes | str = b"\x00", is_trigger: bool = False
    ) -> tuple[int, bytes | None]:
        """
        通过SPI接口发送bytes型数据tx_data，默认返回与tx_data等长的数据

        is_trigger=True 时，tx_data传输开始时 Trigger 信号拉低，完毕后 Trigger 信号拉高

        ::

            TRIG: ───┐            ┌─── HIGH
                     |            |
                     └────────────┘    LOW
                     ┌────────────┐
                     │   tx_data  │
                     └────────────┘
                     ┌────────────┐
                     │   rx_data  │
                     └────────────┘

        is_trigger=False 时，Trigger 信号不变

        ::

            TRIG: ──────────────────── HIGH

                                       LOW
                     ┌────────────┐
                     │   tx_data  │
                     └────────────┘
                     ┌────────────┐
                     │   rx_data  │
                     └────────────┘

        :param tx_data: 要发送数据，bytes或十六进制字符串。
        :type tx_data: str | bytes
        :param rx_count: 接收数据的长度，默认与 tx_data长度一致，如果指定了该长度：
                         1. 发送数据长度小于接收数据长度时，自动在 tx_data后补充dummy数据
                         2. 如果接收数据长度小于发送数据长度，则函数自动把接收到的数据从头截取到 rx_count 长度
        :type rx_count: int
        :param dummy: 发送的填充数据
        :type dummy: bytes | str
        :param is_trigger: 在数据发送时是否产生触发信息（即：发送开始时拉低，结束时拉高）
        :type is_trigger: bool
        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, bytes | None]
        """
        if isinstance(dummy, str):
            dummy = bytes.fromhex(dummy)
        dummy = dummy[:1]
        if isinstance(tx_data, str):
            tx_data = bytes.fromhex(tx_data)
        tx_data_len = len(tx_data)
        if rx_count and rx_count > tx_data_len:
            tx_data += dummy * (rx_count - tx_data_len)

        status, (res, _) = self.spi_transceive_delay_transceive(
            tx_data1=tx_data, tx_data2=None, is_delay=False, delay=0, is_trigger=is_trigger
        )

        if rx_count and tx_data_len > rx_count:
            res = res[:rx_count]

        return status, res

    def i2c_enable(self) -> tuple[int, None]:
        """
        Enable the I2C

        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, None]
        """
        return self._i2c_enable(True)

    def i2c_disable(self) -> tuple[int, None]:
        """
        Disable the I2C

        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, None]
        """
        return self._i2c_enable(False)

    @connection_status_check
    def _i2c_enable(self, enable: bool):
        """
        启用 I2C.

        :param enable: True：启用, False：停用.
        :type enable: bool
        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, None]
        """
        payload = struct.pack(">?", enable)
        self._logger.debug(f"cracker_i2c_enable payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.CRACKER_I2C_ENABLE, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.nut_i2c_enable = enable
        return status, res

    @connection_status_check
    def i2c_reset(self) -> tuple[int, None]:
        """
        重置 I2C 配置。

        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, None]
        """
        payload = None
        self._logger.debug(f"cracker_i2c_reset payload: {payload}")
        return self.send_with_command(protocol.Command.CRACKER_I2C_RESET)

    @connection_status_check
    def i2c_config(
        self, dev_addr: int | None = None, speed: serial.I2cSpeed | None = None, enable_stretch: bool = False
    ) -> tuple[int, None]:
        """
        配置 I2C。

        :param dev_addr: 设备地址.
        :type dev_addr: int
        :param speed: 速度.
        :type speed: serial.I2cSpeed
        :enable_stretch: 是否启用 stretch
        :type enable_stretch: bool
        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, None]
        """

        if dev_addr is None or speed is None:
            config = self.get_current_config()
            if config is None:
                self._logger.error("Get config from cracker error.")
                return self.NON_PROTOCOL_ERROR, None
            if dev_addr is None:
                dev_addr = config.nut_i2c_dev_addr
            if speed is None:
                speed = config.nut_i2c_speed

        payload = struct.pack(">BB?", dev_addr, speed.value, enable_stretch)
        self._logger.debug(f"cracker_i2c_config payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.CRACKER_I2C_CONFIG, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.nut_i2c_dev_addr = dev_addr
            self._config.nut_i2c_speed = speed
        return status, res

    @connection_status_check
    def _i2c_transceive(
        self,
        tx_data: bytes | str | None,
        combined_transfer_count_1: int,
        combined_transfer_count_2: int,
        transfer_rw: tuple[int, int, int, int, int, int, int, int],
        transfer_lens: tuple[int, int, int, int, int, int, int, int],
        is_delay: bool,
        delay: int,
        is_trigger: bool,
    ) -> tuple[int, bytes | None]:
        """
        Basic API for sending and receiving data through the I2C protocol.

        :param tx_data: The data to be sent.
        :type tx_data: bytes | str | None
        :param combined_transfer_count_1: The first combined transmit transfer count.
        :type combined_transfer_count_1: int
        :param combined_transfer_count_2: The second combined transmit transfer count.
        :type combined_transfer_count_2: int
        :param transfer_rw: The read/write configuration tuple of the four transfers in the two sets
                            of Combined Transfer, with a tuple length of 8, where 0 represents write
                            and 1 represents read.
        :type transfer_rw: tuple[int, int, int, int, int, int, int, int, int]
        :param transfer_lens: The transfer length tuple of the four transfers in the two combined transmit sets.
        :type transfer_lens: tuple[int, int, int, int, int, int, int, int]
        :param is_delay: Whether the transmit delay is enabled.
        :type is_delay: bool
        :param delay: Transmit delay duration, in nanoseconds, with a minimum effective duration of 10 nanoseconds.
        :type delay: int
        :param is_trigger: Whether the transmit trigger is enabled.
        :type is_trigger: bool
        :return: The device response status and the data received from the I2C device.
                 Return None if an exception is caught.
        :rtype: tuple[int, bytes | None]
        """
        if isinstance(tx_data, str):
            tx_data = bytes.fromhex(tx_data)

        if combined_transfer_count_1 > 4:
            raise ValueError("Illegal combined combined_transfer_count_1")
        if combined_transfer_count_2 > 4:
            raise ValueError("Illegal combined combined_transfer_count_2")

        if len(transfer_rw) != 8:
            raise ValueError("transfer_rw length must be 8")
        if len(transfer_lens) != 8:
            raise ValueError("transfer_lens length must be 8")

        transfer_rw_num = sum(bit << (7 - i) for i, bit in enumerate(transfer_rw))

        payload = struct.pack(
            ">?I3B8H?",
            is_delay,
            delay,
            combined_transfer_count_1,
            combined_transfer_count_2,
            transfer_rw_num,
            *transfer_lens,
            is_trigger,
        )

        if tx_data is not None:
            payload += tx_data
        status, res = self.send_with_command(protocol.Command.CRACKER_I2C_TRANSCEIVE, payload=payload)
        if status != protocol.STATUS_OK:
            self._logger.error(f"Receive status code error [{status}]")
            return status, None
        else:
            return status, res

    def i2c_transmit(self, tx_data: bytes | str, is_trigger: bool = False) -> tuple[int, None]:
        """
        通过i2c协议发送数据

        is_trigger=True 时，tx_data传输开始时 Trigger 信号拉低，完毕后 Trigger 信号拉高

        is_trigger=True::

            TRIG: ───┐            ┌─── HIGH
                     |            |
                     └────────────┘    LOW
                     ┌────────────┐
                     │   tx_data  │
                     └────────────┘

        is_trigger=False 时，Trigger 信号不变

        is_trigger=False::

            TRIG: ──────────────────── HIGH

                                       LOW
                     ┌────────────┐
                     │   tx_data  │
                     └────────────┘

        :param tx_data: 待发送的数据
        :type tx_data: str | bytes
        :param is_trigger: 在数据发送时是否产生触发信息（即：发送开始时拉低，结束时拉高）
        """
        if isinstance(tx_data, str):
            tx_data = bytes.fromhex(tx_data)
        transfer_rw = (0, 0, 0, 0, 0, 0, 0, 0)
        transfer_lens = (len(tx_data), 0, 0, 0, 0, 0, 0, 0)
        status, _ = self._i2c_transceive(
            tx_data,
            combined_transfer_count_1=1,
            combined_transfer_count_2=0,
            transfer_rw=transfer_rw,
            transfer_lens=transfer_lens,
            is_delay=False,
            delay=0,
            is_trigger=is_trigger,
        )
        return status, None

    def i2c_receive(self, rx_count, is_trigger: bool = False) -> tuple[int, bytes | None]:
        """
        通过i2c协议接收数据

        is_trigger=True 时，rx_data传输开始时 Trigger 信号拉低，完毕后 Trigger 信号拉高

        is_trigger=True::

            TRIG: ───┐            ┌─── HIGH
                     |            |
                     └────────────┘    LOW
                     ┌────────────┐
                     │   rx_data  │
                     └────────────┘

        is_trigger=False 时，Trigger 信号不变

        is_trigger=False::

            TRIG: ──────────────────── HIGH

                                       LOW
                     ┌────────────┐
                     │   rx_data  │
                     └────────────┘

        :param rx_count: 要接收数据的长度
        :type rx_count: int
        :param is_trigger: 在数据接收时是否产生触发信息（即：接收开始时拉低，结束时拉高）
        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, bytes | None]
        """
        transfer_rw = (1, 1, 1, 1, 1, 1, 1, 1)
        transfer_lens = (rx_count, 0, 0, 0, 0, 0, 0, 0)
        return self._i2c_transceive(
            tx_data=None,
            combined_transfer_count_1=1,
            combined_transfer_count_2=0,
            transfer_rw=transfer_rw,
            transfer_lens=transfer_lens,
            is_delay=False,
            delay=0,
            is_trigger=is_trigger,
        )

    def i2c_transmit_delay_receive(
        self, tx_data: bytes | str, delay: int, rx_count: int, is_trigger: bool = False
    ) -> tuple[int, bytes | None]:
        """
        通过i2c协议发送并接收数据

        is_trigger=True 时，tx_data传输开始时 Trigger 信号拉低，完毕后 Trigger 信号拉高
        is_trigger=False是，trigger 信号不变

        is_trigger=True::

            TRIG: ───┐            ┌────────────┐            ┌─── HIGH
                     |            |            |            |
                     └────────────┘            └────────────┘    LOW
                     ┌────────────┬────────────┬────────────┐
                     │  tx_data   │    delay   │   rx_data  │
                     └────────────┴────────────┴────────────┘

        is_trigger=False::

            TRIG: ────────────────────────────────────────────── HIGH

                                                                 LOW
                     ┌────────────┬────────────┬────────────┐
                     │  tx_data   │    delay   │   rx_data  │
                     └────────────┴────────────┴────────────┘

        :param tx_data: 待发送的数据.
        :type tx_data: str | bytes
        :param delay: 发送和接收之间的间隔，单位是10纳秒。
        :type delay: int
        :param rx_count: 要接收的数据长度
        :type rx_count: int
        :param is_trigger: 在数据接收时是否产生触发信息（即：接收开始时拉低，结束时拉高）
        :type is_trigger: bool
        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, bytes | None]
        """
        if isinstance(tx_data, str):
            tx_data = bytes.fromhex(tx_data)
        transfer_rw = (1, 1, 1, 1, 0, 0, 0, 0)
        transfer_lens = (len(tx_data), 0, 0, 0, rx_count, 0, 0, 0)
        return self._i2c_transceive(
            tx_data,
            combined_transfer_count_1=1,
            combined_transfer_count_2=1,
            transfer_rw=transfer_rw,
            transfer_lens=transfer_lens,
            is_delay=True,
            delay=delay,
            is_trigger=is_trigger,
        )

    def i2c_transceive(self, tx_data, rx_count, is_trigger: bool = False) -> tuple[int, bytes | None]:
        """
        通过I2C协议，无延迟的发送并接收数据.

        is_trigger=True 时，tx_data传输开始时 Trigger 信号拉低，完毕后 Trigger 信号拉高
        is_trigger=False是，trigger 信号不变

        is_trigger=True::

            TRIG: ───┐                         ┌─── HIGH
                     |                         |
                     └─────────────────────────┘    LOW
                     ┌────────────┬────────────┐
                     │  tx_data   │   rx_data  │
                     └────────────┴────────────┘

        is_trigger=False::

            TRIG: ──────────────────────────────── HIGH

                                                   LOW
                     ┌────────────┬────────────┐
                     │  tx_data   │   rx_data  │
                     └────────────┴────────────┘

        :param tx_data: 待发送的数据
        :type tx_data: str | bytes
        :param rx_count: 接收数据的长度
        :type rx_count: int
        :param is_trigger: 在数据接收时是否产生触发信息（即：接收开始时拉低，结束时拉高）
        :type is_trigger: bool
        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, bytes | None]
        """
        if isinstance(tx_data, str):
            tx_data = bytes.fromhex(tx_data)
        transfer_rw = (0, 0, 0, 0, 0, 0, 1, 0)
        transfer_lens = (len(tx_data), rx_count, 0, 0, 0, 0, 0, 0)
        return self._i2c_transceive(
            tx_data,
            combined_transfer_count_1=2,
            combined_transfer_count_2=0,
            transfer_rw=transfer_rw,
            transfer_lens=transfer_lens,
            is_delay=False,
            delay=0,
            is_trigger=is_trigger,
        )

    def uart_io_enable(self) -> tuple[int, None]:
        """ "
        使能UART通信接口，使能后TX引脚变为高电平，RX引脚置为三态输入状态。

        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, bytes | None]

        """
        return self._uart_io_enable(True)

    def uart_enable(self) -> tuple[int, None]:
        """
        .. deprecated:: 0.19.0
            此函数将在未来版本中移除，请使用`uart_io_enable`代替。

        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, bytes | None]
        """
        warnings.warn(
            "uart_enable() 已弃用，将在未来版本中移除，请使用 uart_io_enable() 替代。",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return self._uart_io_enable(True)

    def uart_io_disable(self) -> tuple[int, None]:
        """
        关闭UART通信接口，关闭后TX引脚、RX引脚置为三态输入状态。

        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, bytes | None]
        """

        return self._uart_io_enable(False)

    def uart_disable(self) -> tuple[int, None]:
        """
        .. deprecated:: 0.19.0
            此函数将在未来版本中移除，请使用`uart_io_disable`代替。

        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, bytes | None]
        """
        warnings.warn(
            "uart_disable() 已弃用，将在未来版本中移除，请使用 uart_io_disable() 替代。",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return self._uart_io_enable(False)

    @connection_status_check
    def _uart_io_enable(self, enable: bool) -> tuple[int, None]:
        """
        配置UART通信接口使能，
        True: 使能UART通信接口，使能后TX引脚变为高电平，RX引脚置为三态输入状态。
        False: 关闭UART通信接口，关闭后TX引脚、RX引脚置为三态输入状态。

        :param enable: True：使能, False：关闭.
        :type enable: bool
        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, bytes | None]
        """
        payload = struct.pack(">?", enable)
        self._logger.debug(f"cracker_uart_enable payload: {payload.hex()}")
        status, res = self.send_with_command(protocol.Command.CRACKER_UART_ENABLE, payload=payload)
        if status == protocol.STATUS_OK:
            self._config.nut_uart_enable = enable
        return status, res

    @connection_status_check
    def uart_reset(self) -> tuple[int, None]:
        """
        复位UART硬件逻辑。

        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, bytes | None]
        """
        payload = None
        self._logger.debug(f"cracker_uart_reset payload: {payload}")
        return self.send_with_command(protocol.Command.CRACKER_UART_RESET)

    @connection_status_check
    def uart_config(
        self,
        baudrate: serial.Baudrate | None = None,
        bytesize: serial.Bytesize | None = None,
        parity: serial.Parity | None = None,
        stopbits: serial.Stopbits | None = None,
    ) -> tuple[int, None]:
        """
        配置UART接口参数。

        :param baudrate: 波特率。
        :type baudrate: serial.Baudrate
        :param bytesize: 数据位长度。
        :type bytesize: serial.Bytesize
        :param parity: 校验方式。
        :type parity: serial.Parity
        :param stopbits: 停止位。
        :type stopbits: serial.Stopbits
        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, None]
        """

        if baudrate is None or bytesize is None or parity is None or stopbits is None:
            config = self.get_current_config()
            if config is None:
                self._logger.error("Get config from cracker error.")
                return self.NON_PROTOCOL_ERROR, None
            if baudrate is None:
                baudrate = config.nut_uart_baudrate
            if bytesize is None:
                bytesize = config.nut_uart_bytesize
            if parity is None:
                parity = config.nut_uart_parity
            if stopbits is None:
                stopbits = config.nut_uart_stopbits

        payload = struct.pack(">BBBB", stopbits.value, parity.value, bytesize.value, baudrate.value)
        self._logger.debug(f"cracker_uart_config payload: {payload.hex()}")
        return self.send_with_command(protocol.Command.CRACKER_UART_CONFIG, payload=payload)

    @connection_status_check
    def uart_transmit_receive(
        self, tx_data: str | bytes = None, rx_count: int = 0, timeout: int = 10000, is_trigger: bool = False
    ) -> tuple[int, bytes | None]:
        """
        通过UART接口发送数据，等待一段时间后，接收数据，根据用户配置决定是否产生Trigger信号。

        is_trigger=True 时，tx_data传输完毕后，Trigger 信号拉高

        ::

            TRIG: ───┐            ┌────────────┐            ┌─── HIGH
                     |            |            |            |
                     └────────────┘            └────────────┘    LOW
                     ┌────────────┬────────────┬────────────┐
                     │  tx_data   │   process  │   rx_data  │
                     └────────────┴────────────┴────────────┘
                                  | <------ timeout ------> |
        is_trigger=False 时，Trigger 信号不变

        ::

            TRIG: ────────────────────────────────────────────── HIGH

                                                                 LOW
                     ┌────────────┬────────────┬────────────┐
                     │  tx_data   │   process  │   rx_data  │
                     └────────────┴────────────┴────────────┘
                                  | <------ timeout ------> |

        :param tx_data: 要发送数据，bytes或十六进制字符串。
        :type tx_data: str | bytes
        :param rx_count: 要接收数据长度。
        :type rx_count: int
        :param timeout: 超时时间，单位毫秒。
        :type timeout: int
        :param is_trigger: 接收完成时是否产生触发信号。
                           True：接收完成时，Trigger信号拉高，
                           False：接收完成时，Trigger信号不变
        :type is_trigger: bool
        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, bytes | None]
        """
        if isinstance(tx_data, str):
            tx_data = bytes.fromhex(tx_data)

        payload = struct.pack(">H?I", rx_count, is_trigger, timeout)
        if tx_data is not None:
            payload += tx_data
        self._logger.debug(f"cracker_uart_transmit_receive payload: {payload.hex()}")
        return self.send_with_command(protocol.Command.CRACKER_UART_TRANSCEIVE, payload=payload)

    def uart_transmit(self, tx_data: str | bytes, is_trigger: bool = False) -> tuple[int, None]:
        """
        通过UART接口发送数据，根据用户配置决定是否产生Trigger信号。

        is_trigger=True 时，tx_data传输完毕后，Trigger 信号拉高

        ::

            TRIG: ───┐            ┌─── HIGH
                     |            |
                     └────────────┘    LOW
                     ┌────────────┐
                     │  tx_data   │
                     └────────────┘

        is_trigger=False 时，Trigger 信号不变

        ::

            TRIG: ──────────────────── HIGH

                                       LOW
                     ┌────────────┐
                     │  tx_data   │
                     └────────────┘

        :param tx_data: 要发送数据，bytes或十六进制字符串。
        :type tx_data: str | bytes
        :param is_trigger: 接收完成时是否产生触发信号。
                           True：接收完成时，Trigger信号拉高，
                           False：接收完成时，Trigger信号不变
        :type is_trigger: bool
        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, bytes | None]
        """
        return self.uart_transmit_receive(tx_data=tx_data, rx_count=0, timeout=0, is_trigger=is_trigger)

    def uart_receive(self, rx_count: int, timeout: int = 10000, is_trigger: bool = False) -> tuple[int, bytes | None]:
        """
        通过UART接口接收数据，根据用户配置决定是否产生Trigger信号。

        is_trigger=True 时，tx_data传输完毕后，Trigger 信号拉高

        ::

            TRIG: ───┐            ┌─── HIGH
                     |            |
                     └────────────┘    LOW
                     ┌────────────┐
                     │   rx_data  │
                     └────────────┘

        is_trigger=False 时，Trigger 信号不变

        ::

            TRIG: ──────────────────── HIGH

                                       LOW
                     ┌────────────┐
                     │   rx_data  │
                     └────────────┘

        :param rx_count: 要接收数据长度。
        :type rx_count: int
        :param timeout: 超时时间，单位毫秒。
        :type timeout: int
        :param is_trigger: 接收完成时是否产生触发信号。
                           True：接收完成时，Trigger信号拉高，
                           False：接收完成时，Trigger信号不变
        :type is_trigger: bool
        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, bytes | None]
        """
        return self.uart_transmit_receive(tx_data=None, rx_count=rx_count, timeout=timeout, is_trigger=is_trigger)

    @connection_status_check
    def uart_receive_fifo_remained(self) -> tuple[int, int]:
        """
        读取UART接收FIFO剩余未读字节数。

        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, int]
        """
        payload = None
        self._logger.debug(f"cracker_uart_receive_fifo_remained payload: {payload}")
        status, res = self.send_with_command(protocol.Command.CRACKER_UART_RECEIVE_FIFO_REMAINED)
        return status, struct.unpack(">H", res)[0]

    @connection_status_check
    def uart_receive_fifo_dump(self) -> tuple[int, bytes | None]:
        """
        读取UART接收FIFO中剩余的所有数据。

        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, bytes | None]
        """
        payload = None
        self._logger.debug(f"cracker_uart_receive_fifo_dump payload: {payload}")
        return self.send_with_command(protocol.Command.CRACKER_UART_CRACKER_UART_RECEIVE_FIFO_DUMP)

    @connection_status_check
    def uart_receive_fifo_clear(self) -> tuple[int, bytes | None]:
        """
        清除UART接收FIFO中剩余的所有数据。

        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, bytes | None]
        """
        payload = None
        self._logger.debug(f"cracker_uart_receive_fifo_dump payload: {payload}")
        return self.send_with_command(protocol.Command.CRACKER_UART_CRACKER_UART_RECEIVE_CLEAR)

    def nut_reset(self, polar: int = 0, time: int = 10):
        """
        复位nut芯片的RESET管脚，可配置复位电平极性，复位电平时间。默认RESET管脚为三态输入，
        用户在测试板设计时需确保默认不复位（如默认上拉），
        下发nut_reset()命令后，RESET管脚根据极性持续相应的复位时间。

        ::

            polar=LOW
            TRIG: ~~~ ┐            ┌ ~~~
                      └────────────┘     LOW
                      |<-  time  ->|

            polar=HIGH
            TRIG:    ┌────────────┐     HIGH
                 ~~~ ┘            └ ~~~
                     |<-  time  ->|

        :param polar: 极性 0 低电平，1 高电平
        :type polar: int
        :param time: 高低电平持续时间
        :type time: int
        :return: Cracker设备响应状态和接收到的数据：(status, response)。
        :rtype: tuple[int, bytes | None]
        """
        time = time * 1_000_00  # the uint is 10 ns
        payload = struct.pack(">BI", polar, time)
        self._logger.debug(f"cracker_nut_reset payload: {payload}")
        return self.send_with_command(protocol.Command.NUT_RESET, payload=payload)

    def nut_reset_io_enable(self, enable: bool):
        payload = struct.pack(">B", enable)
        self._logger.debug(f"cracker_nut_reset_io_enable payload: {payload}")
        return self.send_with_command(protocol.Command.NUT_RESET_IO_ENABLE, payload=payload)



# Copyright 2024 CrackNuts. All rights reserved.

import pathlib
import typing
from enum import Enum
from typing import Any

from traitlets import traitlets

from cracknuts import logger
from cracknuts.cracker.cracker_s1 import CrackerS1, ConfigS1
from cracknuts.jupyter.panel import MsgHandlerPanelWidget
from cracknuts.jupyter.ui_sync import ConfigProxy, observe_interceptor
import cracknuts.cracker.serial as serial


class CrackerS1PanelWidget(MsgHandlerPanelWidget):
    _esm = pathlib.Path(__file__).parent / "static" / "CrackerS1PanelWidget.js"
    _css = ""

    uri = traitlets.Unicode("cnp://192.168.0.10:8080").tag(sync=True)
    connect_status = traitlets.Bool(False).tag(sync=True)
    cracker_id = traitlets.Unicode("Unknown").tag(sync=True)
    cracker_name = traitlets.Unicode("Unknown").tag(sync=True)
    cracker_version = traitlets.Unicode("Unknown").tag(sync=True)

    # nut
    nut_enable = traitlets.Bool(False).tag(sync=True)
    nut_voltage = traitlets.Float(3.3).tag(sync=True)
    nut_clock_enable = traitlets.Bool(False).tag(sync=True)
    nut_clock = traitlets.Int(65000).tag(sync=True)
    nut_reset_io_enable = traitlets.Bool(False).tag(sync=True)

    nut_uart_enable = traitlets.Bool(False).tag(sync=True)
    nut_uart_baudrate = traitlets.Int(4).tag(sync=True)
    nut_uart_bytesize = traitlets.Int(8).tag(sync=True)
    nut_uart_parity = traitlets.Int(0).tag(sync=True)
    nut_uart_stopbits = traitlets.Int(0).tag(sync=True)

    nut_spi_enable = traitlets.Bool(False).tag(sync=True)
    nut_spi_speed = traitlets.Float(10_000.0).tag(sync=True)
    nut_spi_speed_error = traitlets.Unicode("").tag(sync=True)
    nut_spi_speed_has_error = traitlets.Bool(False).tag(sync=True)
    nut_spi_cpol = traitlets.Int(0).tag(sync=True)
    nut_spi_cpha = traitlets.Int(0).tag(sync=True)
    nut_spi_csn_auto = traitlets.Bool(True).tag(sync=True)
    nut_spi_csn_delay = traitlets.Bool(True).tag(sync=True)

    nut_i2c_enable = traitlets.Bool(False).tag(sync=True)
    nut_i2c_dev_addr = traitlets.Unicode("0x00").tag(sync=True)
    nut_i2c_speed = traitlets.Int(0).tag(sync=True)
    nut_i2c_stretch_enable = traitlets.Bool(False).tag(sync=True)

    nut_timeout = traitlets.Int(0).tag(sync=True)

    # osc
    osc_channel_0_enable = traitlets.Bool(False).tag(sync=True)
    osc_channel_1_enable = traitlets.Bool(True).tag(sync=True)
    sync_sample = traitlets.Bool(False).tag(sync=True)
    sync_args_times = traitlets.Int(1).tag(sync=True)

    osc_sample_clock = traitlets.Int(65000).tag(sync=True)
    osc_sample_phase = traitlets.Int(0).tag(sync=True)
    osc_sample_length = traitlets.Int(1024).tag(sync=True)
    osc_sample_delay = traitlets.Int(1024).tag(sync=True)

    osc_trigger_source = traitlets.Int(0).tag(sync=True)
    osc_trigger_mode = traitlets.Int(0).tag(sync=True)
    osc_trigger_edge = traitlets.Int(0).tag(sync=True)
    osc_trigger_edge_level = traitlets.Int(0).tag(sync=True)

    osc_channel_0_gain = traitlets.Int(1).tag(sync=True)
    osc_channel_1_gain = traitlets.Int(1).tag(sync=True)

    panel_config_different_from_cracker_config = traitlets.Bool(False).tag(sync=True)

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._logger = logger.get_logger(self)
        self.observe: bool = True
        self.cracker: CrackerS1 | None = None
        if "cracker" in kwargs:
            self.cracker: CrackerS1 = kwargs["cracker"]
        if self.cracker is None:
            raise ValueError("cracker is required")
        self.reg_msg_handler("connectButton", "onClick", self.msg_connection_button_on_click)
        self.connect_status = self.cracker.get_connection_status()
        # if self.connect_status:
        # _, self.cracker_id = self.cracker.get_id()
        # _, self.cracker_name = self.cracker.get_hardware_model()
        # _, self.cracker_version = self.cracker.get_firmware_version()
        self._has_connected_before = False

    def read_config_from_cracker(self) -> None:
        # connection
        connect_uri = None
        if self.cracker.get_uri() is not None:
            connect_uri = self.cracker.get_uri()

        cracker_config = self.cracker.get_current_config()
        if cracker_config is None:
            cracker_config = self.cracker.get_default_config().__dict__
        self.update_cracker_panel_config(cracker_config, connect_uri)

    def write_config_to_cracker(self) -> None:
        self.cracker.write_config_to_cracker(self.get_cracker_panel_config())
        self.panel_config_different_from_cracker_config = False
        self.read_config_from_cracker()
        self.listen_cracker_config()

    def get_cracker_panel_config(self) -> ConfigS1:
        panel_config = ConfigS1()
        for prop in panel_config.__dict__.keys():
            default_value = getattr(panel_config, prop)
            if hasattr(self, prop):
                v = getattr(self, prop)
                if prop == "nut_i2c_dev_addr":
                    if v and v.startswith("0x"):
                        v = int(v, base=16)
                    else:
                        v = int(v)
                if default_value is not None and isinstance(default_value, Enum):
                    setattr(panel_config, prop, default_value.__class__(v))
                else:
                    setattr(panel_config, prop, v)
            else:
                self._logger.error(f"Failed to get configuration properties: the widget has no attribute named {prop}")
        return panel_config

    def update_cracker_panel_config(self, config: dict[str, object] | ConfigS1, connect_uri) -> None:
        """
        Sync cracker current to panel(Jupyter widget UI)
        """

        self.observe = False
        self.uri = connect_uri
        if isinstance(config, ConfigS1):
            config = config.__dict__
        for name, value in config.items():
            if value is None:
                continue
            if name.startswith("_"):
                continue
            if hasattr(self, name):
                if isinstance(value, Enum):
                    value = value.value
                if name == "nut_i2c_dev_addr":
                    value = str(value)
                setattr(self, name, value)
            else:
                self._logger.warning(
                    f"Failed to sync configuration to widget: the widget has no attribute named '{name}'."
                )
        self.observe = True

    def listen_cracker_config(self) -> None:
        """
        Bind the cracker and crackerPanel objects so that when the configuration of cracker is set,
        the updated values are automatically synchronized with panel through a ProxyConfig object.
        """
        proxy_config = ConfigProxy(self.cracker.get_current_config(), self)
        self.cracker._config = proxy_config

    def stop_listen_cracker_config(self) -> None:
        config = self.cracker._config
        if isinstance(config, ConfigProxy):
            self.cracker._config = config._config

    def msg_connection_button_on_click(self, args: dict[str, typing.Any]):
        if args.get("action") == "connect":
            self.cracker.connect()
            if self.cracker.get_connection_status():
                self.connect_status = True
                # _, self.cracker_id = self.cracker.get_id()
                # _, self.cracker_name = self.cracker.get_hardware_model()
                # _, self.cracker_version = self.cracker.get_firmware_version()
                self._compare_config(cracker_config=self.cracker.get_current_config())
            else:
                self.connect_status = False
        else:
            self.cracker.disconnect()
            self.connect_status = False
        self.send({"connectFinished": self.connect_status})

    def _compare_config(self, cracker_config: ConfigS1):
        """
        Compare config between current configuration in panel and
        the configuration from cracker after connect from panel.

        """

        if self._has_connected_before:
            panel_cracker_config = self.get_cracker_panel_config().__dict__

            # uri = workspace_config.get("connection")
            if panel_cracker_config:
                for k, v in panel_cracker_config.items():
                    # Skip comparison for ignored configuration items.
                    if k in ("nut_timeout",):
                        continue
                    if hasattr(cracker_config, k):
                        cv = getattr(cracker_config, k)
                        if isinstance(cv, Enum):
                            cv = cv.value
                        if v != cv:
                            self.panel_config_different_from_cracker_config = True
                            self._logger.warning(
                                f"The configuration item {k} differs between the configuration file "
                                f"({v}) and the cracker ({cv})."
                            )
                            break
                    else:
                        self._logger.error(
                            f"Config has no attribute named {k}, "
                            f"which comes from the JSON key in the panel configuration."
                        )
                if not self.panel_config_different_from_cracker_config:
                    self.listen_cracker_config()
            else:
                self._logger.error(
                    "Configuration file format error: The cracker configuration segment is missing. "
                    "The configuration from the cracker or the default configuration will be used."
                )
                self.read_config_from_cracker()
                self.listen_cracker_config()
        else:
            self.read_config_from_cracker()
            self.listen_cracker_config()

    @traitlets.observe("uri")
    @observe_interceptor
    def uri_on_changed(self, change):
        self.cracker.set_uri(change.get("new"))

    @traitlets.observe("nut_enable")
    @observe_interceptor
    def nut_enable_changed(self, change):
        enabled = bool(change.get("new"))
        self.cracker.nut_voltage_enable() if enabled else self.cracker.nut_voltage_disable()
        if enabled:
            self.cracker.nut_voltage(self.cracker.get_current_config().nut_voltage)

    @traitlets.observe("nut_voltage")
    @observe_interceptor
    def nut_voltage_changed(self, change):
        self._logger.error(f"nut voltage changed: {change.get('old')} - {change.get('new')}")
        self._logger.error(f"nut voltage changed: {change}")
        self.cracker.nut_voltage(change.get("new"))

    @traitlets.observe("nut_clock_enable")
    @observe_interceptor
    def nut_clock_enable_changed(self, change):
        enabled = bool(change.get("new"))
        self.cracker.nut_clock_enable() if enabled else self.cracker.nut_clock_disable()
        if enabled:
            self.cracker.nut_clock_freq(self.cracker.get_current_config().nut_clock)

    @traitlets.observe("nut_clock")
    @observe_interceptor
    def nut_clock_changed(self, change):
        self.cracker.nut_clock_freq(int(change.get("new")))

    @traitlets.observe("osc_sample_phase")
    @observe_interceptor
    def osc_sample_phase_changed(self, change):
        self.cracker.osc_sample_clock_phase(int(change.get("new")))

    @traitlets.observe("osc_sample_length")
    @observe_interceptor
    def osc_sample_len_changed(self, change):
        self.cracker.osc_sample_length(int(change.get("new")))

    @traitlets.observe("osc_sample_delay")
    @observe_interceptor
    def osc_sample_delay_changed(self, change):
        self.cracker.osc_sample_delay(int(change.get("new")))

    @traitlets.observe("osc_sample_clock")
    @observe_interceptor
    def osc_sample_rate_changed(self, change):
        self.cracker.osc_sample_clock(int(change.get("new")))

    @traitlets.observe("osc_channel_0_enable")
    @observe_interceptor
    def osc_analog_channel_a_enable_changed(self, change):
        enabled = bool(change.get("new"))
        self.cracker.osc_analog_enable(0) if enabled else self.cracker.osc_analog_disable(0)
        if enabled:
            self.cracker.osc_analog_gain(0, self.cracker.get_current_config().osc_channel_0_gain)

    @traitlets.observe("osc_channel_1_enable")
    @observe_interceptor
    def osc_analog_channel_b_enable_changed(self, change):
        enabled = bool(change.get("new"))
        self.cracker.osc_analog_enable(1) if enabled else self.cracker.osc_analog_disable(1)
        if enabled:
            self.cracker.osc_analog_gain(1, self.cracker.get_current_config().osc_channel_1_gain)

    @traitlets.observe("osc_trigger_source")
    @observe_interceptor
    def osc_trigger_source_changed(self, change):
        self.cracker.osc_trigger_source(change.get("new"))

    @traitlets.observe("osc_trigger_mode")
    @observe_interceptor
    def osc_trigger_mode_changed(self, change):
        self.cracker.osc_trigger_mode(change.get("new"))

    @traitlets.observe("osc_trigger_edge")
    @observe_interceptor
    def osc_trigger_edge_changed(self, change):
        self.cracker.osc_trigger_edge(change.get("new"))

    @traitlets.observe("osc_trigger_edge_level")
    @observe_interceptor
    def osc_trigger_edge_level_changed(self, change):
        self.cracker.osc_trigger_level(change.get("new"))

    @traitlets.observe("osc_channel_0_gain", "osc_channel_1_gain")
    @observe_interceptor
    def osc_analog_channel_gain_changed(self, change):
        name = change.get("name")
        channel = None
        if name == "osc_channel_0_gain":
            channel = 0
        elif name == "osc_channel_1_gain":
            channel = 1
        if channel is not None:
            self.cracker.osc_analog_gain(channel, change.get("new"))

    @traitlets.observe("nut_uart_enable")
    @observe_interceptor
    def nut_uart_enable_changed(self, change):
        enabled = bool(change.get("new"))
        if enabled:
            self.cracker.uart_config(
                serial.Baudrate(self.nut_uart_baudrate),
                serial.Bytesize(self.nut_uart_bytesize),
                serial.Parity(self.nut_uart_parity),
                serial.Stopbits(self.nut_uart_stopbits),
            )

        self.cracker.uart_io_enable() if enabled else self.cracker.uart_io_disable()

    @traitlets.observe("nut_uart_baudrate")
    @observe_interceptor
    def nut_uart_baudrate_changed(self, change):
        self.cracker.uart_config(
            serial.Baudrate(change.get("new")),
            serial.Bytesize(self.nut_uart_bytesize),
            serial.Parity(self.nut_uart_parity),
            serial.Stopbits(self.nut_uart_stopbits),
        )

    @traitlets.observe("nut_uart_bytesize")
    @observe_interceptor
    def nut_uart_bytesize_changed(self, change):
        self.cracker.uart_config(
            serial.Baudrate(self.nut_uart_baudrate),
            serial.Bytesize(change.get("new")),
            serial.Parity(self.nut_uart_parity),
            serial.Stopbits(self.nut_uart_stopbits),
        )

    @traitlets.observe("nut_uart_parity")
    @observe_interceptor
    def nut_uart_parity_changed(self, change):
        self.cracker.uart_config(
            serial.Baudrate(self.nut_uart_baudrate),
            serial.Bytesize(self.nut_uart_bytesize),
            serial.Parity(change.get("new")),
            serial.Stopbits(self.nut_uart_stopbits),
        )

    @traitlets.observe("nut_uart_stopbits")
    @observe_interceptor
    def nut_uart_stopbits_changed(self, change):
        self.cracker.uart_config(
            serial.Baudrate(self.nut_uart_baudrate),
            serial.Bytesize(self.nut_uart_bytesize),
            serial.Parity(self.nut_uart_parity),
            serial.Stopbits(change.get("new")),
        )

    @traitlets.observe("nut_spi_enable")
    @observe_interceptor
    def nut_spi_enable_changed(self, change):
        enabled = bool(change.get("new"))
        if enabled:
            self.cracker.spi_config(
                self.nut_spi_speed,
                serial.SpiCpol(self.nut_spi_cpol),
                serial.SpiCpha(self.nut_spi_cpha),
                self.nut_spi_csn_auto,
                self.nut_spi_csn_delay,
            )
        self.cracker.spi_enable() if enabled else self.cracker.spi_disable()

    @traitlets.observe("nut_spi_speed")
    @observe_interceptor
    def nut_spi_speed_changed(self, change):
        self.nut_spi_speed_has_error = False
        self.nut_spi_speed_error = ""
        status, res = self.cracker.spi_config(
            change.get("new"),
            serial.SpiCpol(self.nut_spi_cpol),
            serial.SpiCpha(self.nut_spi_cpha),
            self.nut_spi_csn_auto,
            self.nut_spi_csn_delay,
        )
        if status != 0:
            if "Not support" in res:
                err_msg = "NotSupportSpeed"
            else:
                err_msg = res
            self.send({"api": "nut_spi_speed", "errorMessage": err_msg, "speed": change.get("new")})

    @traitlets.observe("nut_spi_cpol")
    @observe_interceptor
    def nut_spi_cpol_changed(self, change):
        self.cracker.spi_config(
            self.nut_spi_speed,
            serial.SpiCpol(change.get("new")),
            serial.SpiCpha(self.nut_spi_cpha),
            self.nut_spi_csn_auto,
            self.nut_spi_csn_delay,
        )

    @traitlets.observe("nut_spi_cpha")
    @observe_interceptor
    def nut_spi_cpha_changed(self, change):
        self.cracker.spi_config(
            self.nut_spi_speed,
            serial.SpiCpol(self.nut_spi_cpol),
            serial.SpiCpha(change.get("new")),
            self.nut_spi_csn_auto,
            self.nut_spi_csn_delay,
        )

    @traitlets.observe("nut_spi_csn_auto")
    @observe_interceptor
    def nut_spi_csn_auto_changed(self, change):
        self.cracker.spi_config(
            self.nut_spi_speed,
            serial.SpiCpol(self.nut_spi_cpol),
            serial.SpiCpha(self.nut_spi_cpha),
            change.get("new"),
            self.nut_spi_csn_delay,
        )

    @traitlets.observe("nut_spi_csn_delay")
    @observe_interceptor
    def nut_spi_csn_dly_changed(self, change):
        self.cracker.spi_config(
            self.nut_spi_speed,
            serial.SpiCpol(self.nut_spi_cpol),
            serial.SpiCpha(self.nut_spi_cpha),
            self.nut_spi_csn_auto,
            change.get("new"),
        )

    @traitlets.observe("nut_i2c_enable")
    @observe_interceptor
    def nut_i2c_enable_changed(self, change):
        enabled = bool(change.get("new"))
        if enabled:
            self.cracker.i2c_config(
                int(self.nut_i2c_dev_addr, 16), serial.I2cSpeed(self.nut_i2c_speed), self.nut_i2c_stretch_enable
            )
        self.cracker.i2c_enable() if enabled else self.cracker.i2c_disable()

    @traitlets.observe("nut_i2c_dev_addr")
    @observe_interceptor
    def nut_i2c_dev_addr_changed(self, change):
        self.cracker.i2c_config(
            int(change.get("new"), 16), serial.I2cSpeed(self.nut_i2c_speed), self.nut_i2c_stretch_enable
        )

    @traitlets.observe("nut_i2c_speed")
    @observe_interceptor
    def nut_i2c_speed_changed(self, change):
        self.cracker.i2c_config(
            int(self.nut_i2c_dev_addr, 16), serial.I2cSpeed(change.get("new")), self.nut_i2c_stretch_enable
        )

    @traitlets.observe("nut_i2c_stretch_enable")
    @observe_interceptor
    def nut_i2c_stretch_enable_changed(self, change):
        self.cracker.i2c_config(int(self.nut_i2c_dev_addr, 16), serial.I2cSpeed(self.nut_i2c_speed), change.get("new"))



# Copyright 2024 CrackNuts. All rights reserved.

import os
import pathlib
import typing
from dataclasses import dataclass
from typing import Any
import platform
from cracknuts import logger
from cracknuts.acquisition.acquisition import Acquisition, AcquisitionConfig
from traitlets import traitlets

from cracknuts.jupyter.panel import MsgHandlerPanelWidget


class AcquisitionPanelWidget(MsgHandlerPanelWidget):
    _esm = pathlib.Path(__file__).parent / "static" / "AcquisitionPanelWidget.js"
    _css = ""

    acq_status = traitlets.Int(0).tag(sync=True)
    acq_run_progress = traitlets.Dict({"finished": 0, "total": -1}).tag(sync=True)

    trace_count = traitlets.Int(1000).tag(sync=True)
    sample_offset = traitlets.Int(0).tag(sync=True)
    sample_length = traitlets.Int(1024).tag(sync=True)
    trigger_judge_wait_time = traitlets.Float(0.05).tag(sync=True)
    trigger_judge_timeout = traitlets.Float(0.005).tag(sync=True)
    do_error_max_count = traitlets.Int(1).tag(sync=True)
    file_format = traitlets.Unicode("scarr").tag(sync=True)
    file_path = traitlets.Unicode("").tag(sync=True)
    trace_fetch_interval = traitlets.Float(2.0).tag(sync=True)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._logger = logger.get_logger(self)
        if not hasattr(self, "acquisition"):
            self.acquisition: Acquisition | None = None
            if "acquisition" in kwargs and isinstance(kwargs["acquisition"], Acquisition):
                self.acquisition: Acquisition = kwargs["acquisition"]
            if self.acquisition is None:
                raise ValueError("acquisition is required")

        self.reg_msg_handler("acqStatusButton", "onChange", self.msg_acq_status_changed)
        self.reg_msg_handler("fileSelector", "getDefaultDirectoryInfo", self.msg_get_default_directory_info)
        self.reg_msg_handler("fileSelector", "getDirectoryList", self.msg_get_directory_list)
        self.acquisition.on_status_changed(self.update_acq_status)
        self.acquisition.on_run_progress_changed(self.update_acq_run_progress)
        self.acq_status = self.acquisition.get_status()

    def sync_config_from_acquisition(self) -> None:
        if self.acquisition.trace_count != -1:
            # When acquisition.trace_count = -1, it indicates that the ACQUISITION is in test mode,
            # and in this mode, synchronization of the acquisition trace_count to the GUI is not required.
            self.trace_count = self.acquisition.trace_count
        self.sample_offset = self.acquisition.sample_offset
        self.sample_length = self.acquisition.sample_length
        self.trigger_judge_wait_time = self.acquisition.trigger_judge_wait_time
        self.trigger_judge_timeout = self.acquisition.trigger_judge_timeout
        self.do_error_max_count = self.acquisition.do_error_max_count
        self.file_format = self.acquisition.file_format
        self.file_path = (
            ""
            if self.acquisition.file_path is None or self.acquisition.file_path == "auto"
            else self.acquisition.file_path
        )

    def before_test(self): ...

    def before_run(self): ...

    def listen_acquisition_config(self) -> None:
        ...
        # todo complete ui -> python sync

    def get_acquisition_panel_config(self) -> AcquisitionConfig:
        return AcquisitionConfig(
            self.trace_count,
            self.sample_length,
            self.sample_offset,
            self.trigger_judge_timeout,
            self.trigger_judge_wait_time,
            self.do_error_max_count,
            self.file_path,
            self.file_format,
        )

    def update_acq_status(self, status) -> None:
        self.acq_status = status

    def update_acq_run_progress(self, progress: dict[str, int]):
        self.acq_run_progress = progress

    def msg_acq_status_changed(self, changed: dict[str, typing.Any]):
        status = changed.get("status")
        if status == "pause":
            self.acquisition.pause()
        elif status == "test":
            self.before_test()
            self.acquisition.test(
                sample_offset=self.sample_offset,
                trigger_judge_wait_time=self.trigger_judge_wait_time,
                trigger_judge_timeout=self.trigger_judge_timeout,
                do_error_max_count=self.do_error_max_count,
                trace_fetch_interval=self.trace_fetch_interval,
            )
        elif status == "run":
            self.before_run()
            self.acquisition.run(
                count=self.trace_count,
                sample_offset=self.sample_offset,
                sample_length=self.sample_length,
                trigger_judge_wait_time=self.trigger_judge_wait_time,
                trigger_judge_timeout=self.trigger_judge_timeout,
                do_error_max_count=self.do_error_max_count,
                file_format=self.file_format,
                file_path="auto" if self.file_path == "" or self.file_path is None else self.file_path,
            )
        else:
            self.acquisition.stop()

    def msg_get_default_directory_info(self, args: dict[str, typing.Any]) -> None:
        init_directory = args.get("initDirectory")
        if init_directory is None:
            init_directory = os.path.abspath(os.path.expanduser("~"))
        init_directory = init_directory.replace("\\", "/")
        default_directory_info = _DefaultDirectoryInfo(
            self._get_partitions(), init_directory, self._get_directory_list(init_directory)
        )

        self.send({"defaultDirectoryInfo": default_directory_info.to_dict()})

    def msg_get_directory_list(self, args: dict[str, typing.Any]) -> None:
        path: str = args.get("path")
        self._logger.info(f"get directory list: {path}")
        if path is None:
            return
        path = path.replace("\\", "/")
        directory_list = self._get_directory_list(path)
        self._logger.info(f"directory list: {directory_list}")
        self.send({"directoryList": [d.to_dict() for d in directory_list]})

    @traitlets.observe("trace_fetch_interval")
    def trace_fetch_interval_changed(self, change) -> None:
        if change.get("new"):
            self.acquisition.trace_fetch_interval = change["new"]

    def _get_partitions(self) -> list[str]:
        partitions = []
        system = platform.system()

        if system == "Windows":
            import string
            from ctypes import windll

            try:
                bitmask = windll.kernel32.GetLogicalDrives()
                for letter in string.ascii_uppercase:
                    if bitmask & 1:
                        partitions.append(f"{letter}:/")
                    bitmask >>= 1
            except Exception as e:
                self._logger.error(f"Error getting partitions on Windows: {e}")
        elif system == "Linux":
            try:
                with open("/proc/mounts") as f:
                    for line in f:
                        if line.startswith("/dev/"):
                            parts = line.split()
                            partitions.append(parts[1])
            except Exception as e:
                self._logger.error(f"Error getting partitions on Linux: {e}")

        return partitions

    def _get_directory_list(self, directory: str) -> list["_DirectoryListNode"]:
        directory_list = []
        try:
            path = pathlib.Path(directory)
            if path.is_dir():
                for item in path.iterdir():
                    directory_list.append(_DirectoryListNode(name=item.name, type="folder"))
        except Exception as e:
            self._logger.error(f"Error accessing directory {directory}: {e}")

        return directory_list


@dataclass
class _DirectoryListNode:
    name: str
    type: str

    def to_dict(self):
        return {
            "name": self.name,
            "type": self.type,
        }


@dataclass
class _DefaultDirectoryInfo:
    partitions: list[str]
    path: str
    directory_list: list[_DirectoryListNode]

    def to_dict(self):
        return {
            "partitions": self.partitions,
            "path": self.path,
            "directoryList": [item.to_dict() for item in self.directory_list],
        }



# Copyright 2024 CrackNuts. All rights reserved.

import abc
import json
import os.path
import time
import typing
import warnings

import numpy as np
import zarr

from cracknuts import logger


class TraceDatasetData:
    def __init__(
        self,
        get_trace_data: typing.Callable[[typing.Any, typing.Any], tuple | np.ndarray],
        level: int = 0,
        index: tuple | None = None,
    ):
        self._level: int = level
        self._index: tuple | None = index
        self._get_trace_data = get_trace_data

    def __getitem__(self, index):
        if not isinstance(index, tuple):
            index = (index,)

        level = len(index) + self._level
        index = index if self._index is None else (*self._index, *index)
        if level < 2:
            return TraceDatasetData(self._get_trace_data, level, index)
        else:
            return self._get_trace_data(*index)


class TraceDataset(abc.ABC):
    _channel_names: list[str] | None
    _channel_count: int | None
    _trace_count: int | None
    _sample_count: int | None
    _data_plaintext_length: int | None
    _data_ciphertext_length: int | None
    _data_key_length: int | None
    _data_extended_length: int | None
    _create_time: int | None
    _version: str | None

    @abc.abstractmethod
    def get_origin_data(self): ...

    @classmethod
    @abc.abstractmethod
    def load(cls, path: str, **kwargs) -> "TraceDataset": ...

    @classmethod
    @abc.abstractmethod
    def new(
        cls,
        path: str,
        channel_names: list[str],
        trace_count: int,
        sample_count: int,
        version,
        data_plaintext_length: int | None = None,
        data_ciphertext_length: int | None = None,
        data_key_length: int | None = None,
        data_extended_length: int | None = None,
        **kwargs,
    ) -> "TraceDataset": ...

    @abc.abstractmethod
    def dump(self, path: str | None = None, **kwargs): ...

    @abc.abstractmethod
    def set_trace(
        self,
        channel_name: str,
        trace_index: int,
        trace: np.ndarray,
        data: dict[str, np.ndarray[np.int8] | bytes] | None,
    ): ...

    @property
    def trace_data(self) -> TraceDatasetData:
        """
        曲线及数据（不包含通道、曲线索引信息）
        """
        return TraceDatasetData(get_trace_data=self._get_trace_data)

    @property
    def trace(self):
        """
        曲线数据集中的曲线数据，返回一个支持高级切片索引的对象
        """
        return TraceDatasetData(get_trace_data=self._get_trace)

    @property
    def data(self):
        """
        曲线数据集中的数据，包含明文、密文、扩展数据等，返回一个支持高级切片索引的对象
        """
        return TraceDatasetData(get_trace_data=self._get_data)

    @property
    def trace_data_with_indices(self):
        """
        曲线及数据（包含通道、曲线索引信息）
        """
        return TraceDatasetData(get_trace_data=self._get_trace_data_with_indices)

    def __getitem__(self, item):
        return TraceDatasetData(get_trace_data=self._get_trace_data_with_indices)[item]

    @abc.abstractmethod
    def _get_trace_data_with_indices(
        self, channel_slice, trace_slice
    ) -> tuple[list, list, np.ndarray, list[list[dict[str, bytes | None]]]]: ...

    @abc.abstractmethod
    def _get_trace_data(self, channel_slice, trace_slice) -> tuple[np.ndarray, np.ndarray]: ...

    @abc.abstractmethod
    def _get_trace(self, channel_slice, trace_slice) -> np.ndarray: ...

    @abc.abstractmethod
    def _get_data(self, channel_slice, trace_slice) -> np.ndarray: ...

    @staticmethod
    def _parse_slice(origin_count, index_slice) -> list:
        if origin_count is None:
            raise Exception("origin_count is not set")
        if isinstance(index_slice, slice):
            start, stop, step = index_slice.indices(origin_count)
            indices = [i for i in range(start, stop, step)]
        elif isinstance(index_slice, int):
            indices = [index_slice]
        elif isinstance(index_slice, list):
            indices = index_slice
        else:
            raise ValueError("index_slice is not a slice or list")
        return indices

    def __repr__(self):
        t = type(self)
        return f"<{t.__module__}.{t.__name__} ({self._channel_names}, {self._trace_count})"

    def info(self):
        return _InfoRender(
            self._channel_names,
            self._channel_count,
            self._trace_count,
            self._sample_count,
            self._data_plaintext_length,
            self._data_ciphertext_length,
            self._data_key_length,
            self._data_extended_length,
        )

    @property
    def channel_names(self):
        return self._channel_names

    @property
    def channel_count(self):
        return self._channel_count

    @property
    def trace_count(self):
        return self._trace_count

    @property
    def sample_count(self):
        return self._sample_count

    @property
    def adata_plaintext_length(self):
        return self._data_plaintext_length

    @property
    def data_ciphertext_length(self):
        return self._data_ciphertext_length

    @property
    def data_key_length(self):
        return self._data_key_length

    @property
    def data_extended_length(self):
        return self._data_extended_length

    @property
    def create_time(self):
        return self._create_time


class _InfoRender:
    def __init__(
        self,
        channel_names: list[str],
        channel_count: int,
        trace_count: int,
        sample_count: int,
        data_plaintext_length: int,
        data_ciphertext_length: int,
        data_key_length: int,
        data_extended_length: int,
    ):
        self._channel_names: list[str] = channel_names
        self._channel_count: int = channel_count
        self._trace_count: int = trace_count
        self._sample_count: int = sample_count
        self._data_plaintext_length: int = data_plaintext_length
        self._data_ciphertext_length: int = data_ciphertext_length
        self._data_key_length: int = data_key_length
        self._data_extended_length: int = data_extended_length

    def __repr__(self):
        return (
            f"Channel: {self._channel_names}\r\n"
            f"Trace:   {self._trace_count}, {self._sample_count}\r\n"
            f"Data:    {self._trace_count} "
            f"plaintext: {self._data_plaintext_length} "
            f"ciphertext: {self._data_ciphertext_length} "
            f"key: {self._data_key_length} "
            f"extended: {self._data_extended_length}"
        )


class ZarrTraceDataset(TraceDataset):
    _ATTR_METADATA_KEY = "metadata"
    _GROUP_ROOT_PATH = "0"
    _ARRAY_TRACES_PATH = "traces"
    _ARRAY_DATA_PLAINTEXT_PATH = "plaintext"
    _ARRAY_DATA_CIPHERTEXT_PATH = "ciphertext"
    _ARRAY_DATA_KEY_PATH = "key"
    _ARRAY_DATA_EXTENDED_PATH = "extended"

    def __init__(
        self,
        zarr_path: str,
        create_empty: bool = False,
        channel_names: list[str] | None = None,
        trace_count: int | None = None,
        sample_count: int | None = None,
        data_plaintext_length: int | None = None,
        data_ciphertext_length: int | None = None,
        data_key_length: int | None = None,
        data_extended_length: int | None = None,
        trace_dtype: np.dtype = np.int16,
        zarr_kwargs: dict | None = None,
        zarr_trace_group_kwargs: dict | None = None,
        zarr_data_group_kwargs: dict | None = None,
        create_time: int | None = None,
        version: str | None = None,
    ):
        """
        以Zarr格式存储的 CrackNuts曲线数据集，用户使用时不建议使用构造函数，而是调用 load 函数。
        :param zarr_path: zarr 文件文件路径
        :type zarr_path: str
        :param channel_names: 曲线中通道的名称列表
        :type channel_names: list[str]
        :param trace_count: 曲线条数
        :type trace_count: int
        :param sample_count: 曲线长度（数据点数量）
        :type sample_count: int
        :param data_plaintext_length: 明文长度
        :type data_plaintext_length: int
        :param data_key_length: 密钥长度
        :type data_key_length: int
        :param data_extended_length: 额外数据的长度
        :type data_extended_length: int
        :param trace_dtype: 曲线数据点格式
        :type trace_dtype: np.type
        :param zarr_kwargs: zarr 格式的参数
        :type zarr_kwargs: dict
        :param zarr_trace_group_kwargs: zarr trace group 的参数
        :type zarr_trace_group_kwargs: dict
        :param zarr_data_group_kwargs: zarr data group 的参数
        :type zarr_data_group_kwargs: dict
        :param create_time: 创建时间(unix时间戳)
        :type create_time: int
        :param version: Cracker等版本信息
        :type version: str
        """

        self._zarr_path: str = zarr_path
        self._channel_names: list[str] | None = channel_names
        self._channel_count = None if self._channel_names is None else len(self._channel_names)
        self._trace_count: int | None = trace_count
        self._sample_count: int | None = sample_count
        self._data_plaintext_length: int = data_plaintext_length
        self._data_ciphertext_length: int = data_ciphertext_length
        self._data_key_length: int = data_key_length
        self._data_extended_length: int = data_extended_length
        self._create_time: int | None = create_time
        self._version: str | None = version

        self._logger = logger.get_logger(self)

        if zarr_kwargs is None:
            zarr_kwargs = {}
        if zarr_trace_group_kwargs is None:
            zarr_trace_group_kwargs = {}
        if zarr_data_group_kwargs is None:
            zarr_data_group_kwargs = {}

        mode = zarr_kwargs.pop("mode", "w" if create_empty else "r")
        self._zarr_data = zarr.open(zarr_path, mode=mode, **zarr_kwargs)

        if create_empty:
            if self._channel_names is None or self._trace_count is None or self._sample_count is None:
                raise ValueError(
                    "channel_names and trace_count and sample_count " "must be specified when in write mode."
                )
            self._create_time = int(time.time())
            group_root = self._zarr_data.create_group(self._GROUP_ROOT_PATH)
            for i, _ in enumerate(self._channel_names):
                channel_group = group_root.create_group(str(i))
                channel_group.create(
                    self._ARRAY_TRACES_PATH,
                    shape=(self._trace_count, self._sample_count),
                    dtype=trace_dtype,
                    chunks=(300, sample_count),
                    **zarr_trace_group_kwargs,
                )
                if self._data_plaintext_length is not None:
                    channel_group.create(
                        self._ARRAY_DATA_PLAINTEXT_PATH,
                        shape=(
                            self._trace_count,
                            self._data_plaintext_length,
                        ),
                    )
                if self._data_ciphertext_length is not None:
                    channel_group.create(
                        self._ARRAY_DATA_CIPHERTEXT_PATH,
                        shape=(
                            self._trace_count,
                            self._data_ciphertext_length,
                        ),
                    )
                if self._data_key_length is not None:
                    channel_group.create(
                        self._ARRAY_DATA_KEY_PATH,
                        shape=(
                            self._trace_count,
                            self._data_key_length,
                        ),
                    )
                if self._data_extended_length is not None:
                    channel_group.create(
                        self._ARRAY_DATA_EXTENDED_PATH,
                        shape=(
                            self._trace_count,
                            self._data_extended_length,
                        ),
                    )
            self._zarr_data.attrs[self._ATTR_METADATA_KEY] = {
                "create_time": self._create_time,
                "channel_names": self._channel_names,
                "trace_count": self._trace_count,
                "sample_count": self._sample_count,
                "data_plaintext_length": self._data_plaintext_length,
                "data_ciphertext_length": self._data_ciphertext_length,
                "data_key_length": self._data_key_length,
                "data_extended_length": self._data_extended_length,
                "version": self._version,
            }
        else:
            if self._zarr_path is None:
                raise ValueError("The zarr_path must be specified when in non-write mode.")
            metadata = self._zarr_data.attrs[self._ATTR_METADATA_KEY]
            self._create_time = metadata.get("create_time")
            # This is a piece of logic for handling dataset files compatible with previous versions,
            # which will be removed in subsequent stable versions.
            if "channel_names" not in metadata:
                self._channel_count = metadata.get("channel_count")
                self._channel_names = [str(i) for i in range(self._channel_count)]
            else:
                self._channel_names = metadata.get("channel_names")
                self._channel_count = len(self._channel_names)
            self._trace_count = metadata.get("trace_count")
            self._sample_count = metadata.get("sample_count")
            self._data_plaintext_length = metadata.get("data_plaintext_length")
            self._data_ciphertext_length = metadata.get("data_ciphertext_length")
            self._data_key_length = metadata.get("data_key_length")
            self._data_extended_length = metadata.get("data_extended_length")
            self._version = metadata.get("version")

    @classmethod
    def load(cls, path: str, **kwargs) -> "TraceDataset":
        """
        加载曲线

        :param path: 曲线路径
        :type path: str
        :param kwargs:
        """
        kwargs["mode"] = "r"
        return cls(path, zarr_kwargs=kwargs)

    @classmethod
    def new(
        cls,
        path: str,
        channel_names: list[str],
        trace_count: int,
        sample_count: int,
        version: str,
        data_plaintext_length: int | None = None,
        data_ciphertext_length: int | None = None,
        data_key_length: int | None = None,
        data_extended_length: int | None = None,
        **kwargs,
    ) -> "TraceDataset":
        kwargs["mode"] = "w"
        return cls(
            path,
            create_empty=True,
            channel_names=channel_names,
            trace_count=trace_count,
            sample_count=sample_count,
            version=version,
            data_plaintext_length=data_plaintext_length,
            data_ciphertext_length=data_ciphertext_length,
            data_key_length=data_key_length,
            data_extended_length=data_extended_length,
            zarr_kwargs=kwargs,
        )

    def dump(self, path: str | None = None, **kwargs):
        if path is not None and path != self._zarr_path:
            zarr.copy_store(self._zarr_data, zarr.open(path, mode="w"))

    def set_trace(
        self,
        channel_name: str,
        trace_index: int,
        trace: np.ndarray,
        data: dict[str, np.ndarray[np.int8] | bytes] | None = None,
    ):
        """
        设置曲线，该函数仅需要上位机调用，用户无需调用
        """
        if self._trace_count is None or self._channel_count is None:
            raise Exception("Channel or trace count must has not specified.")
        if channel_name not in self._channel_names:
            raise ValueError("channel index out range")
        if trace_index not in range(0, self._trace_count):
            raise ValueError("trace, index out of range")
        if self._sample_count != trace.shape[0]:
            self._logger.error(
                f"Trace sample count {trace.shape[0]} does not match the previously "
                f"defined value {self._sample_count}, so the trace will be ignored."
            )
            return
        channel_index = self._channel_names.index(channel_name)
        self._get_under_root(channel_index, self._ARRAY_TRACES_PATH)[trace_index] = trace
        if data is not None:
            channel_group = self._get_under_root(str(channel_index))
            for k, v in data.items():
                if isinstance(v, bytes):
                    v = np.frombuffer(v, dtype=np.uint8)
                elif isinstance(v, int):
                    v = np.array([v], dtype=np.uint8)
                data_item_group = channel_group.get(k)
                if data_item_group is None:
                    data_length = v.shape[0]
                    attrs = self._zarr_data.attrs[self._ATTR_METADATA_KEY]
                    if k == "plaintext":
                        self._data_plaintext_length = data_length
                        data_item_group = channel_group.create(
                            k,
                            shape=(
                                self._trace_count,
                                self._data_plaintext_length,
                            ),
                            dtype=np.uint8,
                        )
                        self._zarr_data.attrs[self._ATTR_METADATA_KEY] = attrs | {
                            "data_plaintext_length": self._data_plaintext_length
                        }
                    if k == "ciphertext":
                        self._data_ciphertext_length = data_length
                        data_item_group = channel_group.create(
                            k,
                            shape=(
                                self._trace_count,
                                self._data_ciphertext_length,
                            ),
                            dtype=np.uint8,
                        )
                        self._zarr_data.attrs[self._ATTR_METADATA_KEY] = attrs | {
                            "data_ciphertext_length": self._data_ciphertext_length
                        }
                    if k == "key":
                        self._data_key_length = data_length
                        data_item_group = channel_group.create(
                            k,
                            shape=(
                                self._trace_count,
                                self._data_key_length,
                            ),
                            dtype=np.uint8,
                        )
                        self._zarr_data.attrs[self._ATTR_METADATA_KEY] = attrs | {
                            "data_key_length": self._data_key_length
                        }
                    if k == "extended":
                        self._data_extended_length = data_length
                        data_item_group = channel_group.create(
                            k,
                            shape=(
                                self._trace_count,
                                self._data_extended_length,
                            ),
                            dtype=np.uint8,
                        )
                        self._zarr_data.attrs[self._ATTR_METADATA_KEY] = attrs | {
                            "data_extended_length": self._data_extended_length
                        }
                if data_item_group is not None:
                    data_item_group[trace_index] = v

    def get_origin_data(self) -> zarr.hierarchy.Group:
        """
        获取原始格式的数据，此处返回zarr数据对象

        :return: zarr数据对象
        :rtype: zarr.hierarchy.Group
        """
        return self._zarr_data

    def get_trace_by_indexes(
        self, channel_name: str | int, *trace_indexes: int
    ) -> tuple[np.ndarray, list[dict[str, bytes | None]]] | None:
        """
        根据索引获取曲线数据，该函数用户无需使用
        """
        channel_index = self._channel_names.index(channel_name)
        return (
            self._get_under_root(channel_index, self._ARRAY_TRACES_PATH)[[i for i in trace_indexes]],
            [self._get_data_by_index(channel_index, trace_index) for trace_index in trace_indexes],
        )

    def _get_data_by_index(self, channel_index: int, trace_index: int) -> dict[str, bytes | None]:
        plaintext = self._get_under_root(channel_index, self._ARRAY_DATA_PLAINTEXT_PATH)
        ciphertext = self._get_under_root(channel_index, self._ARRAY_DATA_CIPHERTEXT_PATH)
        key = self._get_under_root(channel_index, self._ARRAY_DATA_KEY_PATH)
        extended = self._get_under_root(channel_index, self._ARRAY_DATA_EXTENDED_PATH)
        if plaintext is not None:
            plaintext = plaintext[trace_index]
            if ciphertext is not None:
                plaintext = plaintext.tobytes()
        if ciphertext is not None:
            ciphertext = ciphertext[trace_index]
            if ciphertext is not None:
                ciphertext = ciphertext.tobytes()
        if key is not None:
            key = key[trace_index]
            if key is not None:
                key = key.tobytes()
        if extended is not None:
            if extended is not None:
                extended = extended.tobytes()
        return {"plaintext": plaintext, "ciphertext": ciphertext, "key": key, "extend": extended}

    def get_trace_by_range(
        self, channel_name: str, index_start: int, index_end: int
    ) -> tuple[np.ndarray, list[dict[str, bytes | None]]] | None:
        """
        根据索引获取曲线数据，该函数用户无需使用
        """
        channel_index = self._channel_names.index(channel_name)
        return (
            self._get_under_root(channel_index, self._ARRAY_TRACES_PATH)[index_start:index_end],
            [self._get_data_by_index(channel_index, trace_index) for trace_index in range(index_start, index_end)],
        )

    def _get_under_root(self, *paths: typing.Any):
        paths = self._GROUP_ROOT_PATH, *paths
        path = "/".join(str(path) for path in paths)
        if path in self._zarr_data:
            return self._zarr_data[path]
        else:
            return None

    def _get_trace_data_with_indices(
        self, channel_slice, trace_slice
    ) -> tuple[list, list, np.ndarray, list[list[dict[str, bytes | None]]]]:
        traces = []
        data = []

        channel_indexes, trace_indexes = (
            self._parse_slice(self._channel_count, channel_slice),
            self._parse_slice(self._trace_count, trace_slice),
        )

        if isinstance(trace_slice, int):
            trace_slice = slice(trace_slice, trace_slice + 1)

        for channel_index in channel_indexes:
            traces.append(self._get_under_root(channel_index, self._ARRAY_TRACES_PATH)[trace_slice])
            data.append([self._get_data_by_index(channel_index, trace_index) for trace_index in trace_indexes])

        return channel_indexes, trace_indexes, np.array(traces), data

    def _get_trace_data(self, channel_slice, trace_slice) -> tuple[np.ndarray, list[list[dict[str, bytes | None]]]]:
        traces = []
        data = []

        channel_indexes, trace_indexes = (
            self._parse_slice(self._channel_count, channel_slice),
            self._parse_slice(self._trace_count, trace_slice),
        )

        if isinstance(trace_slice, int):
            trace_slice = slice(trace_slice, trace_slice + 1)

        for channel_index in channel_indexes:
            traces.append(self._get_under_root(channel_index, self._ARRAY_TRACES_PATH)[trace_slice])
            data.append([self._get_data_by_index(channel_index, trace_index) for trace_index in trace_indexes])

        return np.vstack(traces), data

    def _get_trace(self, channel_slice, trace_slice) -> np.ndarray:
        traces = []

        channel_indexes = self._parse_slice(self.channel_count, channel_slice)

        if isinstance(trace_slice, int):
            trace_slice = slice(trace_slice, trace_slice + 1)

        for channel_index in channel_indexes:
            traces.append(self._get_under_root(channel_index, self._ARRAY_TRACES_PATH)[trace_slice])

        return np.vstack(traces) if len(traces) == 1 else np.stack(traces)

    def _get_data(self, channel_slice, trace_slice) -> list[list[dict[str, bytes | None]]]:
        data = []

        channel_indexes, trace_indexes = (
            self._parse_slice(self._channel_count, channel_slice),
            self._parse_slice(self._trace_count, trace_slice),
        )

        for channel_index in channel_indexes:
            data.append([self._get_data_by_index(channel_index, trace_index) for trace_index in trace_indexes])

        return data


class ScarrTraceDataset(ZarrTraceDataset):
    """
    [DEPRECATED] 这个类已经废弃，请使用 ZarrTraceDataset .
    """

    def __init__(
        self,
        zarr_path: str,
        create_empty: bool = False,
        channel_names: list[str] | None = None,
        trace_count: int | None = None,
        sample_count: int | None = None,
        data_plaintext_length: int | None = None,
        data_ciphertext_length: int | None = None,
        data_key_length: int | None = None,
        data_extended_length: int | None = None,
        trace_dtype: np.dtype = np.int16,
        zarr_kwargs: dict | None = None,
        zarr_trace_group_kwargs: dict | None = None,
        zarr_data_group_kwargs: dict | None = None,
        create_time: int | None = None,
        version: str | None = None,
    ):
        warnings.warn("这个类已经废弃，请使用 ZarrTraceDataset。")
        super().__init__(
            zarr_path,
            create_empty,
            channel_names,
            trace_count,
            sample_count,
            data_plaintext_length,
            data_ciphertext_length,
            data_key_length,
            data_extended_length,
            trace_dtype,
            zarr_kwargs,
            zarr_trace_group_kwargs,
            zarr_data_group_kwargs,
            create_time,
            version,
        )


class NumpyTraceDataset(TraceDataset):
    """
    以numpy格式为基础的TraceDataset
    """

    _ARRAY_TRACE_PATH = "trace.npy"
    _ARRAY_PLAINTEXT_PATH = "plaintext.npy"
    _ARRAY_CIPHERTEXT_PATH = "ciphertext.npy"
    _ARRAY_KEY_PATH = "key.npy"
    _ARRAY_EXTENDED_PATH = "extended.npy"

    _METADATA_PATH = "metadata.json"

    def __init__(
        self,
        path: str | None = None,
        create_empty: bool = False,
        channel_names: list[str] | None = None,
        trace_count: int | None = None,
        sample_count: int | None = None,
        trace_dtype: np.dtype = np.int16,
        data_plaintext_length: int | None = None,
        data_ciphertext_length: int | None = None,
        data_key_length: int | None = None,
        data_extended_length: int | None = None,
        create_time: int | None = None,
        version: str | None = None,
    ):
        self._logger = logger.get_logger(NumpyTraceDataset)

        self._channel_names: list[str] | None = channel_names
        self._channel_count: int | None = None if self._channel_names is None else len(self._channel_names)
        self._trace_count: int | None = trace_count
        self._sample_count: int | None = sample_count
        self._data_plaintext_length: int | None = data_plaintext_length
        self._data_ciphertext_length: int | None = data_ciphertext_length
        self._data_key_length: int | None = data_key_length
        self._data_extended_length: int | None = data_extended_length
        self._create_time: int | None = create_time
        self._version: str | None = version

        self._trace_array: np.ndarray | None = None
        self._plaintext_array: np.ndarray | None = None
        self._ciphertext_array: np.ndarray | None = None
        self._key_array: np.ndarray | None = None
        self._extended_array: np.ndarray | None = None

        if path is not None:
            self._set_path(path)
            self._npy_metadata_path: str = os.path.join(path, self._METADATA_PATH)

        if create_empty:
            if self._channel_names is None or self._trace_count is None or self._sample_count is None:
                raise ValueError(
                    "channel_names and trace_count and sample_count " "must be specified when in write mode."
                )
            self._trace_array = np.zeros(
                shape=(self._channel_count, self._trace_count, self._sample_count), dtype=trace_dtype
            )
            if self._data_plaintext_length is not None:
                self._plaintext_array = np.zeros(
                    shape=(self._channel_count, self._trace_count, self._data_plaintext_length), dtype=np.uint8
                )
            if self._data_ciphertext_length is not None:
                self._ciphertext_array = np.zeros(
                    shape=(self._channel_count, self._trace_count, self._data_ciphertext_length), dtype=np.uint8
                )
            if self._data_key_length is not None:
                self._key_array = np.zeros(
                    shape=(self._channel_count, self._trace_count, self._data_key_length), dtype=np.uint8
                )
            if self._data_extended_length is not None:
                self._extended_array = np.zeros(
                    shape=(self._channel_count, self._trace_count, self._data_extended_length), dtype=np.uint8
                )
            self._create_time = int(time.time())

        else:
            if path is None:
                print("path is required if create_empty is False")
            else:
                self._trace_array = np.load(self._npy_trace_path)

                if not os.path.exists(self._npy_data_plaintext_path):
                    self._logger.warning("npy_data_plaintext_path is not specified, plaintext will be not load.")
                else:
                    self._plaintext_array = np.load(self._npy_data_plaintext_path)

                if not os.path.exists(self._npy_data_ciphertext_path):
                    self._logger.warning("npy_data_ciphertext_path is not specified, ciphertext will be not load.")
                else:
                    self._ciphertext_array = np.load(self._npy_data_ciphertext_path)

                if not os.path.exists(self._npy_data_key_path):
                    self._logger.info("npy_data_key_path is not specified, key will be not load.")
                else:
                    self._key_array = np.load(self._npy_data_key_path)

                if not os.path.exists(self._npy_data_extended_path):
                    self._logger.info("npy_data_extended_path is not specified, extended will be not load.")
                else:
                    self._extended_array = np.load(self._npy_data_extended_path)

                if not os.path.exists(self._npy_metadata_path):
                    self._logger.info("npy_metadata_path is not specified, metadata will be not load.")
                else:
                    self._load_metadata()

    def _load_metadata(self):
        with open(self._npy_metadata_path) as f:
            metadata = json.load(f)
            # This is a piece of logic for handling dataset files compatible with previous versions,
            # which will be removed in subsequent stable versions.
            if "channel_names" not in metadata:
                self._channel_count = metadata["channel_count"]
                self._channel_names = [str(i) for i in range(self._channel_count)]
            else:
                self._channel_names: list[str] | None = metadata.get("channel_names")
                self._channel_count: int | None = len(self._channel_names)
            self._trace_count: int | None = metadata.get("trace_count")
            self._sample_count: int | None = metadata.get("sample_count")
            self._data_plaintext_length: int | None = metadata.get("data_plaintext_length")
            self._data_ciphertext_length: int | None = metadata.get("data_ciphertext_length")
            self._data_key_length: int | None = metadata.get("data_key_length")
            self._data_extended_length: int | None = metadata.get("data_extended_length")
            self._create_time: int | None = metadata.get("create_time")
            self._version: str | None = metadata.get("version")

    def _dump_metadata(self):
        with open(self._npy_metadata_path, "w", encoding="utf-8") as f:
            json.dump(
                {
                    "channel_names": self._channel_names,
                    "trace_count": self._trace_count,
                    "sample_count": self._sample_count,
                    "data_plaintext_length": self._data_plaintext_length,
                    "data_ciphertext_length": self._data_ciphertext_length,
                    "data_key_length": self._data_key_length,
                    "data_extended_length": self._data_extended_length,
                    "create_time": self._create_time,
                    "version": self._version,
                },
                f,
            )

    def get_origin_data(self) -> tuple[np.array, np.array, np.array, np.array, np.array]:
        return self._trace_array, self._plaintext_array, self._ciphertext_array, self._key_array, self._extended_array

    @classmethod
    def load(cls, path: str, **kwargs) -> "TraceDataset":
        return cls(
            path,
            **kwargs,
        )

    @classmethod
    def load_from_numpy_array(cls, trace: np.ndarray):
        channel_count = None
        trace_count = None
        sample_count = None

        shape = trace.shape

        array_size = len(shape)

        if array_size == 1:
            channel_count = 1
            trace_count = 1
            sample_count = shape[0]
        elif array_size == 2:
            channel_count = 1
            trace_count = shape[0]
            sample_count = shape[1]
        elif array_size == 3:
            channel_count = shape[0]
            trace_count = shape[1]
            sample_count = shape[2]

        channel_names = []
        for i in range(channel_count):
            channel_names.append(str(i))

        ds = cls(
            create_empty=True,
            channel_names=channel_names,
            trace_count=trace_count,
            sample_count=sample_count,
            trace_dtype=trace.dtype,
        )

        if array_size == 1:
            ds.set_trace(0, 0, trace, None)
        elif array_size == 2:
            for t in range(shape[0]):
                ds.set_trace(0, t, trace[t], None)
        elif array_size == 3:
            for c in range(shape[0]):
                for t in range(shape[1]):
                    ds.set_trace(c, t, trace[c, t], None)

        return ds

    @classmethod
    def new(
        cls,
        path: str,
        channel_names: list[str],
        trace_count: int,
        sample_count: int,
        version: str,
        data_plaintext_length: int | None = None,
        data_ciphertext_length: int | None = None,
        data_key_length: int | None = None,
        data_extended_length: int | None = None,
        **kwargs,
    ) -> "TraceDataset":
        if not os.path.exists(path):
            os.makedirs(path)
        elif os.path.isfile(path):
            raise Exception(f"{path} is not a file.")

        return cls(
            path=path,
            create_empty=True,
            channel_names=channel_names,
            trace_count=trace_count,
            sample_count=sample_count,
            version=version,
            data_plaintext_length=data_plaintext_length,
            data_ciphertext_length=data_ciphertext_length,
            data_key_length=data_key_length,
            data_extended_length=data_extended_length,
            **kwargs,
        )

    def _set_path(self, path: str):
        self._npy_trace_path: str = os.path.join(path, self._ARRAY_TRACE_PATH)
        self._npy_data_plaintext_path: str = os.path.join(path, self._ARRAY_PLAINTEXT_PATH)
        self._npy_data_ciphertext_path: str = os.path.join(path, self._ARRAY_CIPHERTEXT_PATH)
        self._npy_data_key_path: str = os.path.join(path, self._ARRAY_KEY_PATH)
        self._npy_data_extended_path: str = os.path.join(path, self._ARRAY_EXTENDED_PATH)

    def dump(self, path: str | None = None, **kwargs):
        if path is not None:
            self._set_path(path)
        if self._npy_trace_path is None:
            print("Path must be provided, either as an argument or set in __init__.")
            return
        np.save(self._npy_trace_path, self._trace_array)

        if self._plaintext_array is not None:
            np.save(self._npy_data_plaintext_path, self._plaintext_array)
        if self._ciphertext_array is not None:
            np.save(self._npy_data_ciphertext_path, self._ciphertext_array)
        if self._key_array is not None:
            np.save(self._npy_data_key_path, self._key_array)
        if self._extended_array is not None:
            np.save(self._npy_data_extended_path, self._extended_array)
        self._dump_metadata()

    def set_trace(self, channel_name: str | int, trace_index: int, trace: np.ndarray, data: dict[str, bytes] | None):
        if isinstance(channel_name, int):
            channel_index = channel_name
        else:
            channel_index = self._channel_names.index(channel_name)

        self._trace_array[channel_index, trace_index, :] = trace

        data_plaintext = None if data is None else data.get("plaintext")
        data_ciphertext = None if data is None else data.get("ciphertext")
        data_key = None if data is None else data.get("key")
        data_extended = None if data is None else data.get("extended")

        if data_plaintext is not None:
            if self._plaintext_array is None:
                item_length = len(data_plaintext)
                self._plaintext_array = np.zeros(
                    shape=(self._channel_count, self._trace_count, item_length), dtype=np.uint8
                )
            self._plaintext_array[channel_index, trace_index, :] = np.frombuffer(data_plaintext, dtype=np.uint8)
        if data_ciphertext is not None:
            if self._ciphertext_array is None:
                item_length = len(data_ciphertext)
                self._ciphertext_array = np.zeros(
                    shape=(self._channel_count, self._trace_count, item_length), dtype=np.uint8
                )
            self._ciphertext_array[channel_index, trace_index, :] = np.frombuffer(data_ciphertext, dtype=np.uint8)
        if data_key is not None:
            if self._key_array is None:
                item_length = len(data_key)
                self._key_array = np.zeros(shape=(self._channel_count, self._trace_count, item_length), dtype=np.uint8)
            self._key_array[channel_index, trace_index, :] = np.frombuffer(data_key, dtype=np.uint8)
        if data_extended is not None:
            if self._extended_array is None:
                item_length = len(data_extended)
                self._extended_array = np.zeros(
                    shape=(self._channel_count, self._trace_count, item_length), dtype=np.uint8
                )
            self._key_array[channel_index, trace_index, :] = np.frombuffer(data_extended, dtype=np.uint8)

    def _get_trace_data_with_indices(
        self, channel_slice, trace_slice
    ) -> tuple[list, list, np.ndarray, list[list[dict[str, bytes | None]]]]:
        c = self._parse_slice(self._channel_count, channel_slice)
        t = self._parse_slice(self._trace_count, trace_slice)
        if isinstance(channel_slice, int):
            channel_slice = slice(channel_slice, channel_slice + 1)
        if isinstance(trace_slice, int):
            trace_slice = slice(trace_slice, trace_slice + 1)
        plaintext = None if self._plaintext_array is None else self._plaintext_array[channel_slice, trace_slice]
        ciphertext = None if self._ciphertext_array is None else self._ciphertext_array[channel_slice, trace_slice]
        key = None if self._key_array is None else self._key_array[channel_slice, trace_slice]
        extended = None if self._extended_array is None else self._extended_array[channel_slice, trace_slice]

        print(
            "trace shape",
            self._trace_array.shape,
            self._trace_array[channel_slice, trace_slice].shape,
            channel_slice,
            trace_slice,
        )
        return c, t, self._trace_array[channel_slice, trace_slice], [plaintext, ciphertext, key, extended]

    def _get_trace_data(self, channel_slice, trace_slice) -> tuple[np.ndarray, list[list[dict[str, bytes | None]]]]:
        data = []

        channel_indexes, trace_indexes = (
            self._parse_slice(self._channel_count, channel_slice),
            self._parse_slice(self._trace_count, trace_slice),
        )

        if isinstance(trace_slice, int):
            trace_slice = slice(trace_slice, trace_slice + 1)

        traces = self._trace_array[channel_slice, trace_slice]
        for channel_index in channel_indexes:
            data.append([self._get_data_by_index(channel_index, trace_index) for trace_index in trace_indexes])

        return traces, data

    def _get_trace(self, channel_slice, trace_slice) -> np.ndarray:
        return self._trace_array[channel_slice, trace_slice]

    def _get_data(self, channel_slice, trace_slice) -> list[list[dict[str, bytes | None]]]:
        data = []

        channel_indexes, trace_indexes = (
            self._parse_slice(self._channel_count, channel_slice),
            self._parse_slice(self._trace_count, trace_slice),
        )

        for channel_index in channel_indexes:
            data.append([self._get_data_by_index(channel_index, trace_index) for trace_index in trace_indexes])

        return data

    def _get_data_by_index(self, channel_index: int, trace_index: int) -> dict[str, bytes | None]:
        data = {}
        if self._plaintext_array is not None:
            data["plaintext"] = self._plaintext_array[channel_index, trace_index].tobytes()
        if self._ciphertext_array is not None:
            data["ciphertext"] = self._ciphertext_array[channel_index, trace_index].tobytes()
        if self._key_array is not None:
            data["key"] = self._key_array[channel_index, trace_index].tobytes()
        if self._extended_array is not None:
            data["extended"] = self._extended_array[channel_index, trace_index].tobytes()
        return data



# Copyright 2024 CrackNuts. All rights reserved.

import json
import os
import pathlib
import sys
import typing
from enum import Enum

import traitlets

from cracknuts.cracker.cracker_g1 import CrackerG1
from cracknuts.cracker.cracker_s1 import CrackerS1
from cracknuts.jupyter.acquisition_panel import AcquisitionPanelWidget
from cracknuts.jupyter.cracker_s1_panel import CrackerS1PanelWidget
from cracknuts.jupyter.panel import MsgHandlerPanelWidget
from cracknuts.jupyter.scope_panel import ScopePanelWidget
from cracknuts.utils import user_config


class CracknutsPanelWidget(CrackerS1PanelWidget, AcquisitionPanelWidget, ScopePanelWidget, MsgHandlerPanelWidget):
    _esm = pathlib.Path(__file__).parent / "static" / "CrackNutsPanelWidget.js"
    _css = ""

    cracker_model = traitlets.Unicode("s1").tag(sync=True)
    language = traitlets.Unicode("en").tag(sync=True)

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        if "acquisition" not in kwargs:
            raise ValueError("acquisition must be provided")
        cracker = kwargs["acquisition"].cracker
        kwargs["cracker"] = cracker
        if isinstance(cracker, CrackerS1):
            self.cracker_model = "s1"
        elif isinstance(cracker, CrackerG1):
            self.cracker_model = "g1"
        else:
            raise ValueError(f"can't find panel for cracker type: {type(cracker)}.")
        super().__init__(*args, **kwargs)
        self.language = user_config.get_option("language", fallback="en")

        if self.cracker.get_connection_status():
            self._has_connected_before = True
            workspace_config = self._get_workspace_config()
            if workspace_config:
                config_file_cracker_config = workspace_config.get("cracker")
                # uri = workspace_config.get("connection")
                uri = cracker.get_uri()
                cracker_config = cracker.get_current_config()
                if cracker_config is not None:
                    if config_file_cracker_config:
                        for k, v in config_file_cracker_config.items():
                            # Skip comparison for ignored configuration items.
                            if k in ("nut_timeout",):
                                continue
                            if hasattr(cracker_config, k):
                                cv = getattr(cracker_config, k)
                                if isinstance(cv, Enum):
                                    cv = cv.value
                                if v != cv:
                                    self.panel_config_different_from_cracker_config = True
                                    self._logger.warning(
                                        f"The configuration item {k} differs between the configuration file "
                                        f"({v}) and the cracker ({cv})."
                                    )
                                    break
                            else:
                                self._logger.error(
                                    f"Config has no attribute named {k}, "
                                    f"which comes from the JSON key in the config file."
                                )
                        if not self.panel_config_different_from_cracker_config:
                            self.listen_cracker_config()
                        self.update_cracker_panel_config(config_file_cracker_config, uri)
                    else:
                        self._logger.error(
                            "Configuration file format error: The cracker configuration segment is missing. "
                            "The configuration from the cracker or the default configuration will be used."
                        )
                        self.read_config_from_cracker()
                        self.listen_cracker_config()
                else:
                    self.update_cracker_panel_config(config_file_cracker_config, uri)

                acquisition_config = workspace_config.get("acquisition")
                if acquisition_config:
                    self.acquisition.load_config_from_json(workspace_config.get("acquisition"))
                else:
                    self._logger.error(
                        "Configuration file format error: Acquisition configuration segment is missing. "
                        "The configuration from the acquisition object or "
                        "the default configuration will be used."
                    )
                self.sync_config_from_acquisition()
                self.listen_acquisition_config()
            else:
                self.read_config_from_cracker()
                self.listen_cracker_config()
                self.sync_config_from_acquisition()
                self.listen_acquisition_config()

        self.reg_msg_handler("dumpConfigButton", "onClick", self.dump_config_button_click)
        self.reg_msg_handler("loadConfigButton", "onClick", self.load_config_button_click)
        self.reg_msg_handler("saveConfigButton", "onClick", self.save_config_button_click)
        self.reg_msg_handler("writeConfigButton", "onClick", self.write_config_button_click)
        self.reg_msg_handler("readConfigButton", "onClick", self.read_config_button_click)

    def before_test(self):
        self.write_config_to_cracker()

    def before_run(self):
        self.write_config_to_cracker()

    def dump_config_button_click(self, args: dict[str, typing.Any]):
        self.send({"dumpConfigCompleted": self._dump_config()})

    def load_config_button_click(self, args: dict[str, typing.Any]):
        connection_info = args.get("connection")
        config_file_cracker_config = args.get("cracker")
        acquisition_info = args.get("acquisition")
        cracker_config = self.cracker.get_current_config()
        for k, v in config_file_cracker_config.items():
            if hasattr(cracker_config, k):
                cv = getattr(cracker_config, k)
                if isinstance(cv, Enum):
                    cv = cv.value
                if v != cv:
                    self.panel_config_different_from_cracker_config = True
                    self._logger.error(
                        f"The configuration item {k} differs between the configuration file "
                        f"({v}) and the cracker ({cv})."
                    )
                    break
            else:
                self._logger.error(
                    f"Config has no attribute named {k}, " f"which comes from the JSON key in the config file."
                )

        if self.panel_config_different_from_cracker_config:
            self.stop_listen_cracker_config()
            self.update_cracker_panel_config(config_file_cracker_config, connection_info)

        self.acquisition.load_config_from_str(acquisition_info)
        self.read_config_from_cracker()
        self.send({"loadConfigCompleted": True})

    def write_config_button_click(self, args: dict[str, typing.Any]):
        self.write_config_to_cracker()

    def read_config_button_click(self, args: dict[str, typing.Any]):
        self.read_config_from_cracker()

    def save_config_button_click(self, args: dict[str, typing.Any]):
        current_config_path = self._get_workspace_config_path()
        if current_config_path is None:
            return
        with open(current_config_path, "w") as f:
            self._logger.error(self._dump_config())
            f.write(self._dump_config())

    def _dump_config(self):
        def enum_converter(obj):
            if isinstance(obj, Enum):
                return obj.value
            raise TypeError(f"Type {type(obj)} not serializable")

        return json.dumps(
            {
                "connection": self.uri,
                "cracker": self.get_cracker_panel_config().__dict__,
                "acquisition": self.get_acquisition_panel_config().__dict__,
            },
            indent=4,
            default=enum_converter,
        )

    def _load_current_path_config(self):
        """
        Load the workspace configuration if it exists.
        """
        config = self._get_workspace_config()

        if config is not None:
            connection_info = config.get("connection")
            config_info = config.get("cracker")
            acquisition_info = config.get("acquisition")

            self.cracker.load_config_from_str(config_info)
            self.cracker.set_uri(connection_info)
            self.acquisition.load_config_from_str(acquisition_info)
        else:
            self._logger.debug("Current config path is None or is no exist, current config will be ignored.")

    def _get_workspace_config(self):
        current_config_path = self._get_workspace_config_path()
        if current_config_path is None:
            return
        if current_config_path is not None and os.path.exists(current_config_path):
            with open(current_config_path) as f:
                try:
                    config = json.load(f)
                    return config
                except json.JSONDecodeError as e:
                    self._logger.error(f"Load workspace config file failed: {e.args}")
                    return None

    def _get_workspace_config_path(self):
        global_vars = sys.modules["__main__"].__dict__

        ipynb_path = None

        if "__vsc_ipynb_file__" in global_vars:
            ipynb_path = global_vars["__vsc_ipynb_file__"]
        elif "__session__" in global_vars:
            ipynb_path = global_vars["__session__"]

        if ipynb_path is not None:
            config_path = os.path.join(os.path.dirname(ipynb_path), ".config")
            if not os.path.exists(config_path):
                os.makedirs(config_path)
            elif os.path.isfile(config_path):
                self._logger.error(f"The config directory ({config_path}) already exists, and is not a directory.")
                return None
            return os.path.join(os.path.dirname(ipynb_path), ".config", os.path.basename(ipynb_path)[:-6] + ".json")

    @traitlets.observe("language")
    def on_language_change(self, change):
        user_config.set_option("language", change["new"])
